import logging
import os

import psutil

logger = logging.getLogger(__name__)


def mem_based_cpu_suggestion():
    """
    Return an estimate of the number of clone processes
    that would fit in the currently available memory.
    """

    # Get the resident size of current process
    pid = os.getpid()
    process = psutil.Process(pid)
    rss_mem_usage = process.memory_info().rss

    # Get the free memory
    free_mem = psutil.virtual_memory()[4]

    # Calculate how many processes would fit using a 10% safety margin.
    n = int(free_mem // (rss_mem_usage * 1.1)) + 1  # Count current process

    logger.info(
        f"Current process usage {rss_mem_usage}"
        f" and {free_mem} free memory, could fit {n} processes in RAM."
    )

    return n


def physical_core_cpu_suggestion():
    """
    Return the physical cores.
    """

    n = psutil.cpu_count(logical=False)
    logger.info(f"Found {n} physical cores")
    return n


def virtual_core_cpu_suggestion():
    """
    Return the virtual cores.
    """

    n = psutil.cpu_count(logical=True)
    logger.info(f"Found {n} logical cores")
    return n


def num_procs_suggestion():
    """
    Resolve and return number of processors to use for multiprocessing.

    Query memory and cpu, then makes suggestion.
    """
    suggestions = {
        "memory": mem_based_cpu_suggestion(),
        "physical cores": physical_core_cpu_suggestion(),
    }
    k = sorted(suggestions, key=suggestions.get)[0]
    n = suggestions[k]

    logger.info(f"Suggesting {n} processors based on {k}.")
    return n



import logging

import numpy as np
from numpy import pi
from scipy.special import jv

from aspire.basis import FBBasis2D
from aspire.basis.basis_utils import lgwt
from aspire.nufft import anufft, nufft
from aspire.numeric import fft, xp
from aspire.utils import complex_type
from aspire.utils.matlab_compat import m_reshape

logger = logging.getLogger(__name__)


class FFBBasis2D(FBBasis2D):
    """
    Define a derived class for Fast Fourier Bessel expansion for 2D images

    The expansion coefficients of 2D images on this basis are obtained by
    a fast method instead of the least squares method.
    The algorithm is described in the publication:
    Z. Zhao, Y. Shkolnisky, A. Singer, Fast Steerable Principal Component Analysis,
    IEEE Transactions on Computational Imaging, 2 (1), pp. 1-12 (2016).​

    """

    def _build(self):
        """
        Build the internal data structure to 2D Fourier-Bessel basis
        """
        logger.info(
            "Expanding 2D image in a frequency-domain Fourier–Bessel"
            " basis using the fast method."
        )

        # set cutoff values
        self.rcut = self.nres / 2
        self.kcut = 0.5
        self.n_r = int(np.ceil(4 * self.rcut * self.kcut))
        n_theta = np.ceil(16 * self.kcut * self.rcut)
        self.n_theta = int((n_theta + np.mod(n_theta, 2)) / 2)

        # get upper bound of zeros, ells, and ks  of Bessel functions
        self._calc_k_max()

        # calculate total number of basis functions
        self.count = self.k_max[0] + sum(2 * self.k_max[1:])

        # generate 1D indices for basis functions
        self._compute_indices()
        self._indices = self.indices()

        # get normalized factors
        self.radial_norms, self.angular_norms = self.norms()

        # precompute the basis functions in 2D grids
        self._precomp = self._precomp()

    def _precomp(self):
        """
        Precomute the basis functions on a polar Fourier grid

        Gaussian quadrature points and weights are also generated.
        The sampling criterion requires n_r=4*c*R and n_theta= 16*c*R.

        """
        n_r = self.n_r
        n_theta = self.n_theta
        r, w = lgwt(n_r, 0.0, self.kcut, dtype=self.dtype)

        radial = np.zeros(shape=(np.sum(self.k_max), n_r), dtype=self.dtype)
        ind_radial = 0
        for ell in range(0, self.ell_max + 1):
            for k in range(1, self.k_max[ell] + 1):
                radial[ind_radial] = jv(ell, self.r0[ell][k - 1] * r / self.kcut)
                # NOTE: We need to remove the factor due to the discretization here
                # since it is already included in our quadrature weights
                # Only normalized by the radial part of basis function
                nrm = 1 / (np.sqrt(np.prod(self.sz))) * self.radial_norms[ind_radial]
                radial[ind_radial] /= nrm
                ind_radial += 1

        # Only calculate "positive" frequencies in one half-plane.
        freqs_x = m_reshape(r, (n_r, 1)) @ m_reshape(
            np.cos(np.arange(n_theta, dtype=self.dtype) * 2 * pi / (2 * n_theta)),
            (1, n_theta),
        )
        freqs_y = m_reshape(r, (n_r, 1)) @ m_reshape(
            np.sin(np.arange(n_theta, dtype=self.dtype) * 2 * pi / (2 * n_theta)),
            (1, n_theta),
        )
        freqs = np.vstack((freqs_y[np.newaxis, ...], freqs_x[np.newaxis, ...]))

        return {"gl_nodes": r, "gl_weights": w, "radial": radial, "freqs": freqs}

    def get_radial(self):
        """
        Return precomputed radial part
        """
        return self._precomp["radial"]

    def _evaluate(self, v):
        """
        Evaluate coefficients in standard 2D coordinate basis from those in FB basis

        :param v: A coefficient vector (or an array of coefficient vectors)
            in FB basis to be evaluated. The last dimension must equal `self.count`.
        :return x: The evaluation of the coefficient vector(s) `x` in standard 2D
            coordinate basis. This is Image instance with resolution of `self.sz`
            and the first dimension correspond to remaining dimension of `v`.
        """
        sz_roll = v.shape[:-1]
        v = v.reshape(-1, self.count)

        # number of 2D image samples
        n_data = v.shape[0]

        # get information on polar grids from precomputed data
        n_theta = np.size(self._precomp["freqs"], 2)
        n_r = np.size(self._precomp["freqs"], 1)

        # go through  each basis function and find corresponding coefficient
        pf = np.zeros((n_data, 2 * n_theta, n_r), dtype=complex_type(self.dtype))
        mask = self._indices["ells"] == 0

        ind = 0

        idx = ind + np.arange(self.k_max[0], dtype=int)

        # include the normalization factor of angular part into radial part
        radial_norm = self._precomp["radial"] / np.expand_dims(self.angular_norms, 1)
        pf[:, 0, :] = v[:, mask] @ radial_norm[idx]
        ind = ind + np.size(idx)

        ind_pos = ind

        for ell in range(1, self.ell_max + 1):
            idx = ind + np.arange(self.k_max[ell], dtype=int)
            idx_pos = ind_pos + np.arange(self.k_max[ell], dtype=int)
            idx_neg = idx_pos + self.k_max[ell]

            v_ell = (v[:, idx_pos] - 1j * v[:, idx_neg]) / 2.0

            if np.mod(ell, 2) == 1:
                v_ell = 1j * v_ell

            pf_ell = v_ell @ radial_norm[idx]
            pf[:, ell, :] = pf_ell

            if np.mod(ell, 2) == 0:
                pf[:, 2 * n_theta - ell, :] = pf_ell.conjugate()
            else:
                pf[:, 2 * n_theta - ell, :] = -pf_ell.conjugate()

            ind = ind + np.size(idx)
            ind_pos = ind_pos + 2 * self.k_max[ell]

        # 1D inverse FFT in the degree of polar angle
        pf = 2 * pi * xp.asnumpy(fft.ifft(xp.asarray(pf), axis=1))

        # Only need "positive" frequencies.
        hsize = int(np.size(pf, 1) / 2)
        pf = pf[:, 0:hsize, :]

        for i_r in range(0, n_r):
            pf[..., i_r] = pf[..., i_r] * (
                self._precomp["gl_weights"][i_r] * self._precomp["gl_nodes"][i_r]
            )

        pf = np.reshape(pf, (n_data, n_r * n_theta))

        # perform inverse non-uniformly FFT transform back to 2D coordinate basis
        freqs = m_reshape(self._precomp["freqs"], (2, n_r * n_theta))

        x = 2 * anufft(pf, 2 * pi * freqs, self.sz, real=True)

        # Return X as Image instance with the last two dimensions as *self.sz
        x = x.reshape((*sz_roll, *self.sz))

        return x

    def _evaluate_t(self, x):
        """
        Evaluate coefficient in FB basis from those in standard 2D coordinate basis

        :param x: The Image instance representing coefficient array in the
            standard 2D coordinate basis to be evaluated.
        :return: The evaluation of the coefficient array `x` in the FB basis.
            This is an array of vectors whose last dimension equals `self.count`
            and whose first dimension correspond to `x.shape[0]`.
        """
        # get information on polar grids from precomputed data
        n_theta = np.size(self._precomp["freqs"], 2)
        n_r = np.size(self._precomp["freqs"], 1)
        freqs = np.reshape(self._precomp["freqs"], (2, n_r * n_theta))

        # number of 2D image samples
        n_images = x.shape[0]

        # resamping x in a polar Fourier gird using nonuniform discrete Fourier transform
        pf = nufft(x, 2 * pi * freqs)
        pf = np.reshape(pf, (n_images, n_r, n_theta))

        # Recover "negative" frequencies from "positive" half plane.
        pf = np.concatenate((pf, pf.conjugate()), axis=2)

        # evaluate radial integral using the Gauss-Legendre quadrature rule
        for i_r in range(0, n_r):
            pf[:, i_r, :] = pf[:, i_r, :] * (
                self._precomp["gl_weights"][i_r] * self._precomp["gl_nodes"][i_r]
            )

        #  1D FFT on the angular dimension for each concentric circle
        pf = 2 * pi / (2 * n_theta) * xp.asnumpy(fft.fft(xp.asarray(pf)))

        # This only makes it easier to slice the array later.
        v = np.zeros((n_images, self.count), dtype=x.dtype)

        # go through each basis function and find the corresponding coefficient
        ind = 0
        idx = ind + np.arange(self.k_max[0])
        mask = self._indices["ells"] == 0

        # include the normalization factor of angular part into radial part
        radial_norm = self._precomp["radial"] / np.expand_dims(self.angular_norms, 1)
        v[:, mask] = pf[:, :, 0].real @ radial_norm[idx].T
        ind = ind + np.size(idx)

        ind_pos = ind
        for ell in range(1, self.ell_max + 1):
            idx = ind + np.arange(self.k_max[ell])
            idx_pos = ind_pos + np.arange(self.k_max[ell])
            idx_neg = idx_pos + self.k_max[ell]

            v_ell = pf[:, :, ell] @ radial_norm[idx].T

            if np.mod(ell, 2) == 0:
                v_pos = np.real(v_ell)
                v_neg = -np.imag(v_ell)
            else:
                v_pos = np.imag(v_ell)
                v_neg = np.real(v_ell)

            v[:, idx_pos] = v_pos
            v[:, idx_neg] = v_neg

            ind = ind + np.size(idx)

            ind_pos = ind_pos + 2 * self.k_max[ell]

        return v



"""
ComplexPCA

We're just going to copy scikits PCA and extend to complex.
They refuse to support complex, not because of this code,
but because of how complicated other portions of their
package would become when extended to complex.
They don't want to mix support by admitting complex in only
a few places and not supporting it/crashing in other areas of code.

Unfortunately we need a complex valued PCA, so we wrap theirs for now.
"""

import numpy as np
import scipy.sparse as sp
from sklearn.decomposition import PCA

from .validation import check_array


class ComplexPCA(PCA):
    # In a more ideal world we could patch with something smaller
    # but some versions of scikit do not admit this.
    # def _validate_data(self, *args, **kwargs):
    #     # stuff
    #     return
    # We will instead need to override the method and some dependent methods directly.

    def _fit(self, X):
        """Dispatch to the right submethod depending on the chosen solver."""

        # Raise an error for sparse input.
        # This is more informative than the generic one raised by check_array.
        if sp.issparse(X):
            raise TypeError(
                "PCA does not support sparse input. See "
                "TruncatedSVD for a possible alternative."
            )

        X = check_array(
            X,
            dtype=[np.complex128, np.complex64, np.float64, np.float32],
            ensure_2d=True,
            copy=self.copy,
            allow_complex=True,
        )

        # Handle n_components==None
        if self.n_components is None:
            if self.svd_solver != "arpack":
                n_components = min(X.shape)
            else:
                n_components = min(X.shape) - 1
        else:
            n_components = self.n_components

        # Handle svd_solver
        self._fit_svd_solver = self.svd_solver
        if self._fit_svd_solver == "auto":
            # Small problem or n_components == 'mle', just call full PCA
            if max(X.shape) <= 500 or n_components == "mle":
                self._fit_svd_solver = "full"
            elif n_components >= 1 and n_components < 0.8 * min(X.shape):
                self._fit_svd_solver = "randomized"
            # This is also the case of n_components in (0,1)
            else:
                self._fit_svd_solver = "full"

        # Call different fits for either full or truncated SVD
        if self._fit_svd_solver == "full":
            return self._fit_full(X, n_components)
        elif self._fit_svd_solver in ["arpack", "randomized"]:
            return self._fit_truncated(X, n_components, self._fit_svd_solver)
        else:
            raise ValueError(
                "Unrecognized svd_solver='{0}'" "".format(self._fit_svd_solver)
            )



import logging
import os
import warnings
from concurrent import futures

import mrcfile
import numpy as np
from PIL import Image
from scipy import ndimage, signal
from scipy.ndimage import (
    binary_dilation,
    binary_erosion,
    binary_fill_holes,
    center_of_mass,
)
from sklearn import preprocessing, svm

from aspire.apple.helper import PickerHelper
from aspire.numeric import fft, xp
from aspire.utils import tqdm

logger = logging.getLogger(__name__)


class Picker:
    """
    This class does the actual picking with help
    from PickerHelper class.
    """

    def __init__(
        self,
        particle_size,
        max_size,
        min_size,
        query_size,
        tau1,
        tau2,
        moa,
        container_size,
        filename,
        output_directory,
        model="svm",
        model_opts=None,
        mrc_margin_left=99,
        mrc_margin_right=100,
        mrc_margin_top=99,
        mrc_margin_bottom=100,
        mrc_shrink_factor=2,
        mrc_gauss_filter_size=15,
        mrc_gauss_filter_sigma=0.5,
        response_thresh_norm_factor=20,
        conv_map_nthreads=4,
    ):
        self.particle_size = int(particle_size / 2)
        self.max_size = int(max_size / 2)
        self.min_size = int(min_size / 2)
        self.query_size = int(query_size / 2)
        self.query_size -= self.query_size % 2
        self.tau1 = tau1
        self.tau2 = tau2
        self.moa = int(moa / 2)
        self.container_size = int(container_size / 2)
        self.filename = filename
        self.output_directory = output_directory

        # MRC processing config
        # Margins to discard from any processed .mrc file
        # TODO: Margins are asymmetrical to conform to old behavior - fix going forward
        self.mrc_margin_left = mrc_margin_left
        self.mrc_margin_right = mrc_margin_right
        self.mrc_margin_top = mrc_margin_top
        self.mrc_margin_bottom = mrc_margin_bottom
        self.mrc_shrink_factor = mrc_shrink_factor
        self.mrc_gauss_filter_size = mrc_gauss_filter_size
        self.mrc_gauss_filter_sigma = mrc_gauss_filter_sigma
        self.response_thresh_norm_factor = response_thresh_norm_factor
        self.conv_map_nthreads = conv_map_nthreads

        self.original_im = None  # populated in read_mrc()
        self.im = self.read_mrc()

        if model_opts is None:
            model_opts = dict()
        self.model_opts = model_opts
        self._initialize_model(model)

    def _initialize_model(self, model):
        """
        Util to initialize a classifier model for the Picker object
        based on configuration values.
        """

        logger.info(f"Classifier model desired = {model}")
        if model == "gaussian_mixture":
            from sklearn.mixture import GaussianMixture

            self.model = GaussianMixture(n_components=2)
        elif model == "gaussian_naive_bayes":
            from sklearn.naive_bayes import GaussianNB

            self.model = GaussianNB()
        elif model == "xgboost":
            try:
                import xgboost as xgb
            except ModuleNotFoundError as e:
                logger.error(
                    "`xgboost` import failed. xgboost is an optional package, ensure it is installed."
                )
                raise e

            self.model = xgb.XGBClassifier(
                objective="binary:hinge",
                learning_rate=0.1,
                max_depth=6,
                n_estimators=10,
                method="gpu_hist",
            )
        elif model == "thunder_svm":
            try:
                import thundersvm
            except ModuleNotFoundError as e:
                logger.error(
                    "`thundersvm` import failed. thundersvm is an optional package, ensure it is installed."
                )
                raise e

            self.model = thundersvm.SVC(
                kernel=self.model_opts.get("svm_kernel", "rbf"),
                gamma=float(self.model_opts.get("svm_gamma", 0.5)),
            )
        else:
            logger.info("Using SVM Classifier")

            self.model = svm.SVC(
                C=1,
                kernel=self.model_opts.get("svm_kernel", "rbf"),
                gamma=float(self.model_opts.get("svm_gamma", 0.5)),
                class_weight="balanced",
            )

    def read_mrc(self):
        """
        Reads the micrograph.
        Applies binning and a low-pass filter.

        :return: Micrograph image
        """

        # mrcfile tends to yield many warnings about EMPIAR datasets being corrupt
        # These warnings generally seem benign, and the message could be clearer
        # The following code handles the warnings via ASPIRE's logger
        with warnings.catch_warnings(record=True) as ws:
            # Cause all warnings to always be triggered in this context
            warnings.simplefilter("always")

            # Try to open the mrcfile
            with mrcfile.open(self.filename, mode="r", permissive=True) as mrc:
                im = mrc.data.astype("float")

            # Log each mrcfile warning to debug log, noting the associated file
            for w in ws:
                logger.debug(
                    "In APPLE.picking mrcfile.open reports corruption for"
                    f" {self.filename} warning: {w.message}"
                )

            # Log a single warning to user
            # Give context and note assocated filename
            if len(ws) > 0:
                logger.warning(
                    f"APPLE.picking mrcfile reporting {len(ws)} corruptions."
                    " Most likely this is a problem with the header contents."
                    " Details written to debug log."
                    f" APPLE will attempt to continue processing {self.filename}"
                )

        self.original_im = im

        # Discard outer pixels
        im = im[
            self.mrc_margin_top : -self.mrc_margin_bottom,
            self.mrc_margin_left : -self.mrc_margin_right,
        ]

        # Make square
        side_length = min(im.shape)
        im = im[:side_length, :side_length]

        size = tuple((np.array(im.shape) / self.mrc_shrink_factor).astype(int))

        # Note, float64 required for signal.correlate call accuracy.
        im = np.asarray(
            Image.fromarray(im).resize(size, Image.Resampling.BICUBIC)
        ).astype(np.float64, copy=False)

        im = signal.correlate(
            im,
            PickerHelper.gaussian_filter(
                self.mrc_gauss_filter_size, self.mrc_gauss_filter_sigma
            ),
            "same",
        )

        return im

    def query_score(self):
        """
        Calculates score for each query image.

        Extracts query images and reference windows.
        Computes the cross-correlation between these windows,
        and applies a threshold to compute a score for each query image.

        :return: Matrix containing a score for each query image.
        """

        micro_img = xp.asarray(self.im)
        logger.info("Extracting query images")
        query_box = PickerHelper.extract_query(micro_img, self.query_size // 2)
        logger.info("Extracting query images complete")

        query_box = xp.conj(fft.fft2(query_box, axes=(2, 3)))

        reference_box = PickerHelper.extract_references(
            micro_img, self.query_size, self.container_size
        )

        reference_size = PickerHelper.reference_size(micro_img, self.container_size)
        conv_map = xp.zeros((reference_size, query_box.shape[0], query_box.shape[1]))

        def _work(index):
            reference_box_i = fft.fft2(reference_box[index], axes=(0, 1))
            window_t = xp.multiply(reference_box_i, query_box)
            cc = fft.ifft2(window_t, axes=(2, 3))
            return index, cc.real.max((2, 3)) - cc.real.mean((2, 3))

        n_works = reference_size
        pbar = tqdm(total=reference_size)

        # Ideally we'd like something like 'SerialExecutor' to enable easy debugging
        # but for now do an if-else
        if self.conv_map_nthreads > 1:
            with futures.ThreadPoolExecutor(self.conv_map_nthreads) as executor:
                to_do = [executor.submit(_work, i) for i in range(n_works)]

                for future in futures.as_completed(to_do):
                    i, res = future.result()
                    conv_map[i, :, :] = res
                    pbar.update(1)
        else:
            for i in range(n_works):
                _, conv_map[i, :, :] = _work(i)
                pbar.update(1)

        pbar.close()

        conv_map = xp.transpose(conv_map, (1, 2, 0))

        min_val = xp.min(conv_map)
        max_val = xp.max(conv_map)
        thresh = min_val + (max_val - min_val) / self.response_thresh_norm_factor
        return xp.asnumpy(xp.sum(conv_map >= thresh, axis=2))

    def run_svm(self, score):
        """
        Trains and uses an SVM classifier.

        Trains an SVM classifier to distinguish between noise
        and particle projections based on mean intensity and variance.
        Every possible window in the micrograph is then classified as either noise or particle.

        This results in a segmentation of the micrograph.

        :param score: Matrix containing a score for each query image.
        :return: Segmentation of the micrograph into noise and particle projections.
        """

        micro_img = xp.asarray(self.im)
        particle_windows = np.floor(self.tau1)
        non_noise_windows = np.ceil(self.tau2)
        bw_mask_p, bw_mask_n = Picker.get_maps(
            self, score, micro_img, particle_windows, non_noise_windows
        )

        x, y = PickerHelper.get_training_set(
            micro_img, bw_mask_p, bw_mask_n, self.query_size
        )
        x = xp.asnumpy(x)
        y = xp.asnumpy(y)

        scaler = preprocessing.StandardScaler()
        scaler.fit(x)
        x = scaler.transform(x)
        classifier = self.model
        classifier.fit(x, y)

        mean_all, std_all = PickerHelper.moments(micro_img, self.query_size)
        mean_all = xp.asnumpy(mean_all)
        std_all = xp.asnumpy(std_all)

        mean_all = mean_all[
            self.query_size - 1 : -(self.query_size - 1),
            self.query_size - 1 : -(self.query_size - 1),
        ]

        std_all = std_all[
            self.query_size - 1 : -(self.query_size - 1),
            self.query_size - 1 : -(self.query_size - 1),
        ]

        mean_all = np.reshape(mean_all, (np.prod(mean_all.shape), 1), "F")
        std_all = np.reshape(std_all, (np.prod(std_all.shape), 1), "F")
        cls_input = np.concatenate((mean_all, std_all), axis=1)
        cls_input = scaler.transform(cls_input)

        # compute classification for all possible windows in micrograph
        segmentation = classifier.predict(cls_input)

        _segmentation_shape = int(np.sqrt(segmentation.shape[0]))
        segmentation = np.reshape(
            segmentation, (_segmentation_shape, _segmentation_shape), "F"
        )

        return segmentation.copy()

    def morphology_ops(self, segmentation):
        """
        Discards suspected artifacts from segmentation.

        :param segmentation: Segmentation of the micrograph into noise and particle projections.

        :return: Updated segmentation of the micrograph into noise and particle projections.
        """

        if (binary_fill_holes(segmentation) == np.ones(segmentation.shape)).all():
            segmentation[0:100, 0:100] = np.zeros((100, 100))

        segmentation = binary_fill_holes(segmentation)
        y, x = np.ogrid[
            -self.min_size : self.min_size + 1, -self.min_size : self.min_size + 1
        ]
        element = x * x + y * y <= self.min_size * self.min_size
        segmentation_e = binary_erosion(segmentation, element)

        y, x = np.ogrid[
            -self.max_size : self.max_size + 1, -self.max_size : self.max_size + 1
        ]
        element = x * x + y * y <= self.max_size * self.max_size
        segmentation_o = binary_erosion(segmentation, element)
        segmentation_o = np.reshape(
            segmentation_o, (segmentation_o.shape[0], segmentation_o.shape[1], 1), "F"
        )

        size_const, _ = ndimage.label(segmentation_e, np.ones((3, 3)))
        size_const = np.reshape(
            size_const, (size_const.shape[0], size_const.shape[1], 1), "F"
        )
        labels = np.unique(size_const * segmentation_o)
        idx = np.where(labels != 0)
        labels = np.take(labels, idx)
        labels = np.reshape(labels, (1, 1, np.prod(labels.shape)), "F")

        matrix1 = np.repeat(size_const, labels.shape[2], 2)
        matrix2 = np.repeat(labels, matrix1.shape[0], 0)
        matrix2 = np.repeat(matrix2, matrix1.shape[1], 1)

        matrix3 = np.equal(matrix1, matrix2)
        matrix4 = np.sum(matrix3, 2)

        segmentation_e[np.where(matrix4 == 1)] = 0

        return segmentation_e

    def extract_particles(self, segmentation):
        """
        Saves particle centers into output .star file,
        after dismissing regions that are too big to
        contain a particle.

        :param segmentation: Segmentation of the micrograph into noise and particle projections.
        :return: centers

        """
        segmentation = segmentation[
            self.query_size // 2 - 1 : -self.query_size // 2,
            self.query_size // 2 - 1 : -self.query_size // 2,
        ]
        labeled_segments, _ = ndimage.label(segmentation, np.ones((3, 3)))
        values, repeats = np.unique(labeled_segments, return_counts=True)

        values_to_remove = np.where(repeats > self.max_size**2)
        values = np.take(values, values_to_remove)
        values = np.reshape(values, (1, 1, np.prod(values.shape)), "F")

        labeled_segments = np.reshape(
            labeled_segments,
            (labeled_segments.shape[0], labeled_segments.shape[1], 1),
            "F",
        )
        matrix1 = np.repeat(labeled_segments, values.shape[2], 2)
        matrix2 = np.repeat(values, matrix1.shape[0], 0)
        matrix2 = np.repeat(matrix2, matrix1.shape[1], 1)

        matrix3 = np.equal(matrix1, matrix2)
        matrix4 = np.sum(matrix3, 2)

        segmentation[np.where(matrix4 == 1)] = 0
        labeled_segments, _ = ndimage.label(segmentation, np.ones((3, 3)))

        max_val = np.amax(
            np.reshape(labeled_segments, (np.prod(labeled_segments.shape)))
        )
        center = center_of_mass(segmentation, labeled_segments, np.arange(1, max_val))
        center = np.rint(center)

        if len(center) == 0:
            logger.warning(
                "Picker did not find any centers.  "
                "Review micrograph, particle_size, min, and max particle sizes."
            )
            return center

        img = np.zeros((segmentation.shape[0], segmentation.shape[1]))
        img[center[:, 0].astype(int), center[:, 1].astype(int)] = 1
        y, x = np.ogrid[-self.moa : self.moa + 1, -self.moa : self.moa + 1]
        element = x * x + y * y <= self.moa * self.moa
        img = binary_dilation(img, structure=element)
        labeled_img, _ = ndimage.label(img, np.ones((3, 3)))
        values, repeats = np.unique(labeled_img, return_counts=True)
        y = np.where(repeats == np.count_nonzero(element))
        y = np.array(y)
        y = y.astype(int)
        y = np.reshape(y, (np.prod(y.shape)), "F")
        y -= 1
        center = center[y, :]

        center = center + (self.query_size // 2 - 1) * np.ones(center.shape)
        center = center + (self.query_size // 2 - 1) * np.ones(center.shape)
        center = center + np.ones(center.shape)

        center = self.mrc_shrink_factor * center

        # swap columns to align with Relion
        center = center[:, [1, 0]]

        # first column is x; second column is y - offset by margins that were discarded from the image
        center[:, 0] += self.mrc_margin_left
        center[:, 1] += self.mrc_margin_top

        if self.output_directory is not None:
            basename = os.path.basename(self.filename)
            name_str, ext = os.path.splitext(basename)

            applepick_path = os.path.join(
                self.output_directory, "{}_applepick.star".format(name_str)
            )
            with open(applepick_path, "w") as f:
                np.savetxt(
                    f,
                    ["data_root\n\nloop_\n_rlnCoordinateX #1\n_rlnCoordinateY #2"],
                    fmt="%s",
                )
                np.savetxt(f, center, fmt="%d %d")

        return center

    def get_maps(self, score, micro_img, particle_windows, non_noise_windows):
        """
        Gets maps of regions from which to extract particle
        training for the SVM classifier.

        :param score: Matrix containing a score for each query image.
        :param micro_img: Micrograph image.
        :param particle_windows: Number of windows that must contain a particle.
        :param non_noise_windows: Number of windows that contain neither noise nor particles.
        :return: 2-Tuple of particle and noise masks
        """
        particles = particle_windows.astype(int)
        non_noise = non_noise_windows.astype(int)

        idx = np.argsort(-np.reshape(score, (np.prod(score.shape)), "F"))
        x, y = np.unravel_index(idx, score.shape)
        bw_mask_p = np.zeros((micro_img.shape[0], micro_img.shape[1]))
        qs = self.query_size // 2

        begin_row_idx = y * qs
        end_row_idx = np.minimum(begin_row_idx + self.query_size, bw_mask_p.shape[0])
        begin_col_idx = x * qs
        end_col_idx = np.minimum(begin_col_idx + self.query_size, bw_mask_p.shape[1])

        for j in range(particles):
            bw_mask_p[
                begin_row_idx[j] : end_row_idx[j], begin_col_idx[j] : end_col_idx[j]
            ] = 1

        bw_mask_n = np.copy(bw_mask_p)
        for j in range(particles, non_noise):
            bw_mask_n[
                begin_row_idx[j] : end_row_idx[j], begin_col_idx[j] : end_col_idx[j]
            ] = 1

        return bw_mask_p, bw_mask_n



import logging
from collections.abc import Iterable

import numpy as np

from aspire.basis import Basis
from aspire.utils import complex_type

logger = logging.getLogger(__name__)


class SteerableBasis2D(Basis):
    """
    SteerableBasis2D is an extension of Basis that is expected to have
    `rotation` (steerable) and `calculate_bispectrum` methods.
    """

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        # Because they are used for core features of SteerableBasis2D,
        #   cache the indices for positive and negative ells.
        # Note zero is special case.
        self._zero_angular_inds = self.angular_indices == 0
        self._pos_angular_inds = (self.signs_indices == 1) & (self.angular_indices != 0)
        self._neg_angular_inds = self.signs_indices == -1

    def calculate_bispectrum(
        self, complex_coef, flatten=False, filter_nonzero_freqs=False, freq_cutoff=None
    ):
        """
        Calculate bispectrum for a set of coefs in this basis.

        The Bispectum matrix is of shape:
            (count, count, unique_radial_indices)

        where count is the number of complex coefficients.

        :param coef: Coefficients representing a (single) image expanded in this basis.
        :param flatten: Optionally extract symmetric values (tril) and then flatten.
        :param filter_nonzero_freqs: Remove indices corresponding to zero frequency (defaults False).
        :param freq_cutoff: Truncate (zero) high k frequecies above (int) value, defaults off (None).
        :return: Bispectum matrix (complex valued).
        """

        # Check shape
        if complex_coef.shape[0] != 1:
            raise ValueError(
                "Due to potentially large sizes, bispectrum is limited to a single set of coefs."
                f"  Passed shape {complex_coef.shape}"
            )

        if complex_coef.shape[1] != self.complex_count:
            raise ValueError(
                "Basis.calculate_bispectrum coefs expected"
                f" to have (complex) count {self.complex_count}, received {complex_coef.shape}."
            )

        # From here just treat complex_coef as 1d vector instead of 1 by count 2d array.
        complex_coef = complex_coef[0]

        if freq_cutoff and freq_cutoff > np.max(self.complex_angular_indices):
            logger.warning(
                f"Bispectrum frequency cutoff {freq_cutoff} outside max {np.max(self.complex_angular_indices)}"
            )

        # Notes, regarding the naming:
        # radial freq indices q in paper/slides, _indices["ks"] in code
        radial_indices = self.complex_radial_indices  # q
        # angular freq indices k in paper/slides, _indices["ells"] in code
        angular_indices = self.complex_angular_indices  # k
        # Compute the set of all unique q in the compressed basis
        #   Note that np.unique is also sorted.
        unique_radial_indices = np.unique(radial_indices)

        # When compressed, we need maps between the basis and uncompressed set of q
        #   to a reduced set of q that remain after compression.
        # One map is provided by self.complex_radial_indices
        #   which maps an index in the basis remaining after compression to a q value.
        # The following code computes a similar but inverted map,
        #   given a q value, find an index into the set of unique q after compression.
        # Also, it is known that the set of q gets sparser with increasing k
        #   but that's ignored here, instead construct a dense
        #   array and filter it later.
        #  The plan is to revisit this code after appropriate coef classes are derived.

        # Default array to fill_value, we can use a value
        # k should never achieve..
        fill_value = self.complex_count**2
        compressed_radial_map = (
            np.ones(np.max(unique_radial_indices) + 1, dtype=int) * fill_value
        )
        for uniq_q_index, q_value in enumerate(unique_radial_indices):
            compressed_radial_map[q_value] = uniq_q_index

        B = np.zeros(
            (self.complex_count, self.complex_count, unique_radial_indices.shape[0]),
            dtype=complex_type(self.dtype),
        )

        logger.info(f"Calculating bispectrum matrix with shape {B.shape}.")

        for ind1 in range(self.complex_count):
            k1 = angular_indices[ind1]
            if freq_cutoff and k1 > freq_cutoff:
                continue
            coef1 = complex_coef[ind1]

            for ind2 in range(self.complex_count):
                k2 = angular_indices[ind2]
                if freq_cutoff and k2 > freq_cutoff:
                    continue
                coef2 = complex_coef[ind2]

                k3 = k1 + k2
                intermodulated_coef_inds = angular_indices == k3

                if np.any(intermodulated_coef_inds):
                    # Get the specific q indices related to feasible k3 angular_indices
                    Q3_ind = radial_indices[intermodulated_coef_inds]

                    if hasattr(self, "compressed") and self.compressed:
                        # Map those Q3_ind values to indices into compressed unique_radial_indices
                        #  by using the compressed_radial_map prepared above.
                        Q3_ind = compressed_radial_map[Q3_ind]

                    Coef3 = complex_coef[intermodulated_coef_inds]

                    B[ind1, ind2, Q3_ind] = coef1 * coef2 * np.conj(Coef3)

        if filter_nonzero_freqs:
            non_zero_freqs = angular_indices != 0
            B = B[non_zero_freqs][:, non_zero_freqs]

        if flatten:
            # B is sym, start by taking lower triangle.
            tril = np.tri(B.shape[0], dtype=bool)
            B = B[tril, :]
            # Then flatten
            B = B.flatten()

        return B

    def rotate(self, coef, radians, refl=None):
        """
        Returns coefs rotated counter-clockwise by `radians`.

        :param coef: Basis coefs.
        :param radians: Rotation in radians.
        :param refl: Optional reflect image (about y=0) (bool)
        :return: rotated coefs.
        """

        # Enforce a stack axis to support sanity checks
        coef = np.atleast_2d(coef)

        # Covert radians to a broadcastable shape
        if isinstance(radians, Iterable):
            radians = np.fromiter(radians, dtype=self.dtype).reshape(-1, 1)
            if len(radians) != len(coef):
                raise RuntimeError(
                    "`rotate` call `radians` length cannot broadcast with"
                    f" `coef` {len(coef)} != {len(radians)}"
                )
        # else: radians can be a constant

        assert (
            self.count == coef.shape[-1]
        ), "Number of coefficients must match self.count."

        # self.angular_indices are `ks`
        # For all coef in stack,
        #   compute the ks * radian used in the trig functions
        ks_rad = np.atleast_2d(self.angular_indices * radians)
        ks_pos = ks_rad[:, self._pos_angular_inds]
        ks_neg = ks_rad[:, self._neg_angular_inds]

        # Slice the coef on postive and negative ells
        coef_zer = coef[:, self._zero_angular_inds]
        coef_pos = coef[:, self._pos_angular_inds]
        coef_neg = coef[:, self._neg_angular_inds]

        # Handle zero case and avoid mutating the original array
        coef = np.empty_like(coef)
        coef[:, self._zero_angular_inds] = coef_zer

        # refl
        if refl is not None:
            if isinstance(refl, np.ndarray):
                assert len(refl) == len(coef)
            # else: refl can be a constant
            # negate the coefs corresponding to negative ells
            coef_neg[refl] = coef_neg[refl] * -1

        # Apply formula
        coef[:, self._pos_angular_inds] = coef_pos * np.cos(ks_pos) + coef_neg * np.sin(
            ks_neg
        )
        coef[:, self._neg_angular_inds] = coef_neg * np.cos(ks_neg) - coef_pos * np.sin(
            ks_pos
        )

        return coef

    def complex_rotate(self, complex_coef, radians, refl=None):
        """
        Returns complex coefs rotated counter-clockwise by `radians`.

        This implementation uses the complex exponential.
        It is kept in the code for documentation and
        reference purposes.

        To invoke in code:

        self.to_real(
            self.complex_rotate(
                self.to_complex(coef), radians, refl)
            )
        )

        :param complex_coef: Basis coefs (in complex representation).
        :param radians: Rotation in radians.
        :param refl: Optional reflect image (about y=0) (bool)
        :return: rotated (complex) coefs.
        """

        # Covert radians to a broadcastable shape
        if isinstance(radians, np.ndarray):
            if len(radians) != len(complex_coef):
                raise RuntimeError(
                    "`rotate` call `radians` length cannot broadcast with"
                    f" `complex_coef` {len(complex_coef)} != {len(radians)}"
                )
            radians = radians.reshape(-1, 1)
        # else: radians can be a constant

        ks = self.complex_angular_indices
        assert len(ks) == complex_coef.shape[-1]

        # Don't mutate the input coef array (danger)
        _complex_coef = complex_coef.copy()

        # refl
        if refl is not None:
            if isinstance(refl, np.ndarray):
                assert len(refl) == len(complex_coef)
            # else: refl can be a constant
            # get the coefs corresponding to -ks , aka "ells"
            _complex_coef[refl] = np.conj(complex_coef[refl])

        _complex_coef = _complex_coef * np.exp(1j * ks * radians)

        return _complex_coef

    def shift(self, coef, shifts):
        """
        Returns coefs shifted by `shifts`.

        This will transform to real cartesian space, shift,
        and transform back to Polar Fourier-Bessel space.

        :param coef: Basis coefs.
        :param shifts: Shifts in pixels (x,y). Shape (1,2) or (len(coef), 2).
        :return: coefs of shifted images.
        """

        shifts = np.atleast_2d(np.array(shifts))
        if shifts.ndim != 2:
            raise ValueError("`shifts` should be a one or two dimensional array.")
        if shifts.shape[1] != 2 or shifts.shape[0] not in (1, len(coef)):
            raise ValueError(
                "`shifts` should be shape (1,2) or (len(coef),2),"
                f" received {shifts.shape}."
            )

        return self.evaluate_t(self.evaluate(coef).shift(shifts))



import logging
import math

import numpy as np
import scipy.sparse as sparse

from aspire.abinitio.orientation_src import OrientEstSource
from aspire.basis import PolarBasis2D
from aspire.utils import common_line_from_rots
from aspire.utils.random import choice

logger = logging.getLogger(__name__)


class CLOrient3D:
    """
    Define a base class for estimating 3D orientations using common lines methods
    """

    def __init__(
        self, src, n_rad=None, n_theta=360, n_check=None, max_shift=0.15, shift_step=1
    ):
        """
        Initialize an object for estimating 3D orientations using common lines

        :param src: The source object of 2D denoised or class-averaged imag
        :param n_rad: The number of points in the radial direction. If None,
            n_rad will default to the ceiling of half the resolution of the source.
        :param n_theta: The number of points in the theta direction. This value must be even.
            Default is 360.
        :param n_check: For each image/projection find its common-lines with
            n_check images. If n_check is less than the total number of images,
            a random subset of n_check images is used.
        :param max_shift: Determines maximum range for shifts as a proportion
            of the resolution. Default is 0.15.
        :param shift_step: Resolution of shift estimation in pixels.
            Default is 1 pixel.
        """
        self.src = src
        # Note dtype is inferred from self.src
        self.dtype = self.src.dtype
        self.n_img = src.n
        self.n_res = self.src.L
        self.n_rad = n_rad
        self.n_theta = n_theta
        self.n_check = n_check
        self.clmatrix = None
        self.max_shift = math.ceil(max_shift * self.n_res)
        self.shift_step = shift_step
        self.rotations = None

        self._build()

    def _build(self):
        """
        Build the internal data structure for orientation estimation
        """
        if self.n_rad is None:
            self.n_rad = math.ceil(0.5 * self.n_res)
        if self.n_check is None:
            self.n_check = self.n_img
        if self.n_theta % 2 == 1:
            msg = "n_theta must be even"
            logger.error(msg)
            raise NotImplementedError(msg)

        imgs = self.src.images[:]

        # Obtain coefficients in polar Fourier basis for input 2D images
        self.basis = PolarBasis2D(
            (self.n_res, self.n_res), self.n_rad, self.n_theta, dtype=self.dtype
        )
        self.pf = self.basis.evaluate_t(imgs)
        self.pf = self.pf.reshape(self.n_img, self.n_theta, self.n_rad)

        n_theta_half = self.n_theta // 2

        # The last two dimension of pf is of size n_theta x n_rad. We will convert pf
        # into an array of size (n_theta/2) x (n_rad-1), that is, take half of each ray
        # through the origin except the DC part, and also take the angles only up to PI.
        # This is due to the fact that the original images are real, and thus each ray
        # is conjugate symmetric. We therefore gain nothing by taking longer correlations
        # (of length 2*n_rad-1 instead of n_rad-1). In the Matlab version, pf is converted to
        # the size of (n_theta/2) x (2*n_rad-1) but most of the calculations of build_clmatrix
        # and estimate_shifts below only use the size of (n_theta/2) x (n_rad-1). In the
        # Python version we will use the size of (n_theta/2) x (n_rad-1) directly and make
        # sure every part is using it. By taking shorter correlations we can speed the
        # computation by a factor of two.
        self.pf = np.flip(self.pf[:, n_theta_half:, 1:], 2)

    def estimate_rotations(self):
        """
        Estimate orientation matrices for all 2D images

        Subclasses should implement this function.
        """
        raise NotImplementedError("subclasses should implement this")

    def save_rotations(self):
        """
        Return a new ImageSource object for orientation information
        """
        return OrientEstSource(self.src, self)

    def build_clmatrix(self):
        """
        Build common-lines matrix from Fourier stack of 2D images
        """

        n_img = self.n_img
        n_check = self.n_check

        if self.n_theta % 2 == 1:
            msg = "n_theta must be even"
            logger.error(msg)
            raise NotImplementedError(msg)

        n_theta_half = self.n_theta // 2

        # need to do a copy to prevent modifying self.pf for other functions
        pf = self.pf.copy()

        # Allocate local variables for return
        # clmatrix represents the common lines matrix.
        # Namely, clmatrix[i,j] contains the index in image i of
        # the common line with image j. Note the common line index
        # starts from 0 instead of 1 as Matlab version. -1 means
        # there is no common line such as clmatrix[i,i].
        clmatrix = -np.ones((n_img, n_img), dtype=self.dtype)
        # When cl_dist[i, j] is not -1, it stores the maximum value
        # of correlation between image i and j for all possible 1D shifts.
        # We will use cl_dist[i, j] = -1 (including j<=i) to
        # represent that there is no need to check common line
        # between i and j. Since it is symmetric,
        # only above the diagonal entries are necessary.
        cl_dist = -np.ones((n_img, n_img), dtype=self.dtype)

        # Allocate variables used for shift estimation

        # set maximum value of 1D shift (in pixels) to search
        # between common-lines.
        max_shift = self.max_shift
        # Set resolution of shift estimation in pixels. Note that
        # shift_step can be any positive real number.
        shift_step = self.shift_step
        # 1D shift between common-lines
        shifts_1d = np.zeros((n_img, n_img))

        # Prepare the shift phases to try and generate filter for common-line detection
        r_max = pf.shape[2]
        shifts, shift_phases, h = self._generate_shift_phase_and_filter(
            r_max, max_shift, shift_step
        )
        all_shift_phases = shift_phases.T

        # Apply bandpass filter, normalize each ray of each image
        # Note that only use half of each ray
        pf = self._apply_filter_and_norm("ijk, k -> ijk", pf, r_max, h)

        # Search for common lines between [i, j] pairs of images.
        # Creating pf and building common lines are different to the Matlab version.
        # The random selection is implemented.
        for i in range(n_img - 1):
            p1 = pf[i]
            p1_real = np.real(p1)
            p1_imag = np.imag(p1)

            # build the subset of j images if n_check < n_img
            n_remaining = n_img - i - 1
            n_j = min(n_remaining, n_check)
            subset_j = np.sort(choice(n_remaining, n_j, replace=False) + i + 1)

            for j in subset_j:
                p2_flipped = np.conj(pf[j])

                for shift in range(len(shifts)):
                    shift_phases = all_shift_phases[shift]
                    p2_shifted_flipped = (shift_phases * p2_flipped).T
                    # Compute correlations in the positive r direction
                    part1 = p1_real.dot(np.real(p2_shifted_flipped))
                    # Compute correlations in the negative r direction
                    part2 = p1_imag.dot(np.imag(p2_shifted_flipped))

                    c1 = part1 - part2
                    sidx = c1.argmax()
                    cl1, cl2 = np.unravel_index(sidx, c1.shape)
                    sval = c1[cl1, cl2]

                    c2 = part1 + part2
                    sidx = c2.argmax()
                    cl1_2, cl2_2 = np.unravel_index(sidx, c2.shape)
                    sval2 = c2[cl1_2, cl2_2]

                    if sval2 > sval:
                        cl1 = cl1_2
                        cl2 = cl2_2 + n_theta_half
                        sval = sval2
                    sval = 2 * sval
                    if sval > cl_dist[i, j]:
                        clmatrix[i, j] = cl1
                        clmatrix[j, i] = cl2
                        cl_dist[i, j] = sval
                        shifts_1d[i, j] = shifts[shift]

        self.clmatrix = clmatrix
        self.shifts_1d = shifts_1d

    def estimate_shifts(self, equations_factor=1, max_memory=4000):
        """
        Estimate 2D shifts in images

        This function computes 2D shifts in x, y of images by solving the least-squares
        equations to `Ax = b`. `A` on the left-hand side is a sparse matrix representing
        precomputed coefficients of shift equations; and on the right-side, `b` is
        estimated 1D shifts along the theta direction between two Fourier rays (one in
        image i and the other in image j).  Each row of shift equations contains four
        unknowns, shifts in x, y for a pair of images. The detailed implementation
        can be found in the book chapter as below:
        Y. Shkolnisky and A. Singer,
        Center of Mass Operators for Cryo-EM - Theory and Implementation,
        Modeling Nanoscale Imaging in Electron Microscopy,
        T. Vogt, W. Dahmen, and P. Binev (Eds.)
        Nanostructure Science and Technology Series,
        Springer, 2012, pp. 147–177

        :param equations_factor: The factor to rescale the number of shift equations
            (=1 in default)
        :param max_memory: If there are N images and N_check selected to check
            for common lines, then the exact system of equations solved for the shifts
            is of size 2N x N(N_check-1)/2 (2N unknowns and N(N_check-1)/2 equations).
            This may be too big if N is large. The algorithm will use `equations_factor`
            times the total number of equations if the resulting total number of memory
            requirements is less than `max_memory` (in megabytes); otherwise it will
            reduce the number of equations by approximation to fit in `max_memory`.
        """

        # Generate approximated shift equations from estimated rotations
        shift_equations, shift_b = self._get_shift_equations_approx(
            equations_factor, max_memory
        )

        # Solve the linear equation, optionally printing numerical debug details.
        show = False
        if logging.getLogger().isEnabledFor(logging.DEBUG):
            show = True
        # Negative sign comes from using -i conversion of Fourier transformation
        est_shifts = sparse.linalg.lsqr(shift_equations, -shift_b, show=show)[0]
        est_shifts = est_shifts.reshape((2, self.n_img), order="F")

        return est_shifts

    def _get_shift_equations_approx(self, equations_factor=1, max_memory=4000):
        """
        Generate approximated shift equations from estimated rotations

        The function computes the common lines from the estimated rotations,
        and then, for each common line, estimates the 1D shift between its two
        Fourier rays (one in image i and the other in image j). Using the common
        lines and the 1D shifts, shift equations are generated randomly based
        on a memory factor and represented by sparse matrix.

        This function processes the (Fourier transformed) images exactly as the
        `build_clmatrix` function.

        :param equations_factor: The factor to rescale the number of shift equations
            (=1 in default)
        :param max_memory: If there are N images and N_check selected to check
            for common lines, then the exact system of equations solved for the shifts
            is of size 2N x N(N_check-1)/2 (2N unknowns and N(N_check-1)/2 equations).
            This may be too big if N is large. The algorithm will use `equations_factor`
            times the total number of equations if the resulting total number of
            memory requirements is less than `max_memory` (in megabytes); otherwise it
            will reduce the number of equations to fit in `max_memory`.

        :return; The left and right-hand side of shift equations
        """

        n_theta_half = self.n_theta // 2
        n_img = self.n_img
        rotations = self.rotations

        pf = self.pf.copy()

        # Estimate number of equations that will be used to calculate the shifts
        n_equations = self._estimate_num_shift_equations(
            n_img, equations_factor, max_memory
        )
        # Allocate local variables for estimating 2D shifts based on the estimated number
        # of equations. The shift equations are represented using a sparse matrix,
        # since each row in the system contains four non-zeros (as it involves
        # exactly four unknowns). The variables below are used to construct
        # this sparse system. The k'th non-zero element of the equations matrix
        # is stored at index (shift_i(k),shift_j(k)).
        shift_i = np.zeros(4 * n_equations, dtype=self.dtype)
        shift_j = np.zeros(4 * n_equations, dtype=self.dtype)
        shift_eq = np.zeros(4 * n_equations, dtype=self.dtype)
        shift_b = np.zeros(n_equations, dtype=self.dtype)

        # Prepare the shift phases to try and generate filter for common-line detection
        # The shift phases are pre-defined in a range of max_shift that can be
        # applied to maximize the common line calculation. The common-line filter
        # is also applied to the radial direction for easier detection.
        max_shift = self.max_shift
        shift_step = self.shift_step
        r_max = pf.shape[2]
        _, shift_phases, h = self._generate_shift_phase_and_filter(
            r_max, max_shift, shift_step
        )

        d_theta = np.pi / n_theta_half

        # Generate two index lists for [i, j] pairs of images
        idx_i, idx_j = self._generate_index_pairs(n_equations)

        # Go through all shift equations in the size of n_equations
        # Iterate over the common lines pairs and for each pair find the 1D
        # relative shift between the two Fourier lines in the pair.
        for shift_eq_idx in range(n_equations):
            i = idx_i[shift_eq_idx]
            j = idx_j[shift_eq_idx]
            # get the common line indices based on the rotations from i and j images
            c_ij, c_ji = self._get_cl_indices(rotations, i, j, n_theta_half)

            # Extract the Fourier rays that correspond to the common line
            pf_i = pf[i, c_ij]

            # Check whether need to flip or not Fourier ray of j image
            # Is the common line in image j in the positive
            # direction of the ray (is_pf_j_flipped=False) or in the
            # negative direction (is_pf_j_flipped=True).
            is_pf_j_flipped = c_ji >= n_theta_half
            if not is_pf_j_flipped:
                pf_j = pf[j, c_ji]
            else:
                pf_j = pf[j, c_ji - n_theta_half]

            # perform bandpass filter, normalize each ray of each image,
            pf_i = self._apply_filter_and_norm("i, i -> i", pf_i, r_max, h)
            pf_j = self._apply_filter_and_norm("i, i -> i", pf_j, r_max, h)

            # apply the shifts to images
            pf_i_flipped = np.conj(pf_i)
            pf_i_stack = np.einsum("i, ij -> ij", pf_i, shift_phases)
            pf_i_flipped_stack = np.einsum("i, ij -> ij", pf_i_flipped, shift_phases)

            c1 = 2 * np.real(np.dot(np.conj(pf_i_stack.T), pf_j))
            c2 = 2 * np.real(np.dot(np.conj(pf_i_flipped_stack.T), pf_j))

            # find the indices for the maximum values
            # and apply corresponding shifts
            sidx1 = np.argmax(c1)
            sidx2 = np.argmax(c2)
            sidx = sidx1 if c1[sidx1] > c2[sidx2] else sidx2
            dx = -max_shift + sidx * shift_step

            # Create a shift equation for the image pair [i,j]
            idx = np.arange(4 * shift_eq_idx, 4 * shift_eq_idx + 4)
            # angle of common ray in image i
            shift_alpha = c_ij * d_theta
            # Angle of common ray in image j.
            shift_beta = c_ji * d_theta
            # Row index to construct the sparse equations
            shift_i[idx] = shift_eq_idx
            # Columns of the shift variables that correspond to the current pair [i, j]
            shift_j[idx] = [2 * i, 2 * i + 1, 2 * j, 2 * j + 1]
            # Right hand side of the current equation
            shift_b[shift_eq_idx] = dx

            # Compute the coefficients of the current equation
            coeffs = np.array(
                [
                    np.sin(shift_alpha),
                    np.cos(shift_alpha),
                    -np.sin(shift_beta),
                    -np.cos(shift_beta),
                ]
            )
            shift_eq[idx] = -1 * coeffs if is_pf_j_flipped else coeffs

        # create sparse matrix object only containing non-zero elements
        shift_equations = sparse.csr_matrix(
            (shift_eq, (shift_i, shift_j)),
            shape=(n_equations, 2 * n_img),
            dtype=self.dtype,
        )

        return shift_equations, shift_b

    def _estimate_num_shift_equations(self, n_img, equations_factor=1, max_memory=4000):
        """
        Estimate total number of shift equations in images

        The function computes total number of shift equations based on
        number of images and preselected memory factor.
        :param n_img:  The total number of input images
        :param equations_factor: The factor to rescale the number of shift equations
            (=1 in default)
        :param max_memory: If there are N images and N_check selected to check
            for common lines, then the exact system of equations solved for the shifts
            is of size 2N x N(N_check-1)/2 (2N unknowns and N(N_check-1)/2 equations).
            This may be too big if N is large. The algorithm will use `equations_factor`
            times the total number of equations if the resulting total number of
            memory requirements is less than `max_memory` (in megabytes); otherwise it
            will reduce the number of equations to fit in `max_memory`.
        :return: Estimated number of shift equations
        """
        # Number of equations that will be used to estimation the shifts
        n_equations_total = int(np.ceil(n_img * (self.n_check - 1) / 2))
        # Estimated memory requirements for the full system of equation.
        # This ignores the sparsity of the system, since backslash seems to
        # ignore it.
        memory_total = equations_factor * (
            n_equations_total * 2 * n_img * self.dtype.itemsize
        )
        if memory_total < (max_memory * 10**6):
            n_equations = int(np.ceil(equations_factor * n_equations_total))
        else:
            subsampling_factor = (max_memory * 10**6) / memory_total
            subsampling_factor = min(1.0, subsampling_factor)
            n_equations = int(np.ceil(n_equations_total * subsampling_factor))

        if n_equations < n_img:
            logger.warning(
                "Too few equations. Increase memory_factor. Setting n_equations to n_img."
            )
            n_equations = n_img

        if n_equations < 2 * n_img:
            logger.warning(
                "Number of equations is small. Consider increase memory_factor."
            )

        return n_equations

    def _generate_shift_phase_and_filter(self, r_max, max_shift, shift_step):
        """
        Prepare the shift phases and generate filter for common-line detection

        The shift phases are pre-defined in a range of max_shift that can be
        applied to maximize the common line calculation. The common-line filter
        is also applied to the radial direction for easier detection.
        :param r_max: Maximum index for common line detection
        :param max_shift: Maximum value of 1D shift (in pixels) to search
        :param shift_step: Resolution of shift estimation in pixels
        :return: shift phases matrix and common lines filter
        """

        # Number of shifts to try
        n_shifts = int(np.ceil(2 * max_shift / shift_step + 1))

        # only half of ray
        rk = np.arange(-r_max, 0)

        # Generate all shift phases
        shifts = -max_shift + shift_step * np.arange(n_shifts)
        shift_phases = np.exp(np.outer(-2 * np.pi * 1j * rk / (2 * r_max + 1), shifts))
        # Set filter for common-line detection
        h = np.sqrt(np.abs(rk)) * np.exp(-np.square(rk) / (2 * (r_max / 4) ** 2))
        return shifts, shift_phases, h

    def _generate_index_pairs(self, n_equations):
        """
        Generate two index lists for [i, j] pairs of images
        """
        idx_i = []
        idx_j = []
        for i in range(self.n_img - 1):
            tmp_j = range(i + 1, self.n_img)
            idx_i.extend([i] * len(tmp_j))
            idx_j.extend(tmp_j)
        idx_i = np.array(idx_i, dtype="int")
        idx_j = np.array(idx_j, dtype="int")

        # Select random pairs based on the size of n_equations
        rp = choice(np.arange(len(idx_j)), size=n_equations, replace=False)

        return idx_i[rp], idx_j[rp]

    def _get_cl_indices(self, rotations, i, j, n_theta):
        """
        Get common line indices based on the rotations from i and j images

        :param rotations: Array of rotation matrices
        :param i: Index for i image
        :param j: Index for j image
        :param n_theta: Total number of common lines
        :return: Common line indices for i and j images
        """
        # get the common line indices based on the rotations from i and j images
        r_i = rotations[i]
        r_j = rotations[j]
        c_ij, c_ji = common_line_from_rots(r_i.T, r_j.T, 2 * n_theta)

        # To match clmatrix, c_ij is always less than PI
        # and c_ji may be be larger than PI.
        if c_ij >= n_theta:
            c_ij -= n_theta
            c_ji -= n_theta
        if c_ji < 0:
            c_ji += 2 * n_theta

        return c_ij, c_ji

    def _apply_filter_and_norm(self, subscripts, pf, r_max, h):
        """
        Apply common line filter and normalize each ray

        :subscripts: Specifies the subscripts for summation of Numpy
            `einsum` function
        :param pf: Fourier transform of images
        :param r_max: Maximum index for common line detection
        :param h: common lines filter
        :return: filtered and normalized i images
        """

        # Note if we'd rather not have the dtype and casting args,
        #   we can control h.dtype instead.
        np.einsum(subscripts, pf, h, out=pf, dtype=pf.dtype, casting="same_kind")
        pf[..., r_max - 1 : r_max + 2] = 0
        pf /= np.linalg.norm(pf, axis=-1)[..., np.newaxis]

        return pf



import logging
from functools import partial

import numpy as np
import scipy.sparse.linalg
from scipy.linalg import norm
from scipy.sparse.linalg import LinearOperator

from aspire.reconstruction.kernel import FourierKernel
from aspire.volume import Volume

logger = logging.getLogger(__name__)


class Estimator:
    def __init__(self, src, basis, batch_size=512, preconditioner="circulant"):
        """
        An object representing a 2*L-by-2*L-by-2*L array containing the non-centered Fourier transform of the mean
        least-squares estimator kernel.
        Convolving a volume with this kernel is equal to projecting and backproject-ing that volume in each of the
        projection directions (with the appropriate amplitude multipliers and CTFs) and averaging over the whole
        dataset.
        Note that this is a non-centered Fourier transform, so the zero frequency is found at index 1.
        """

        self.src = src
        self.basis = basis
        self.dtype = self.src.dtype
        self.batch_size = batch_size
        self.preconditioner = preconditioner

        if not self.dtype == self.basis.dtype:
            logger.warning(
                f"Inconsistent types in {self.dtype} Estimator."
                f" basis: {self.basis.dtype}"
            )

        if src.L != basis.nres:
            raise ValueError(
                "Currently require 2D source and 3D volume resolution to be the same."
                f" Given src.L={src.L} != {basis.nres}"
            )

    def __getattr__(self, name):
        """Lazy attributes instantiated on first-access"""

        if name == "kernel":
            logger.info("Computing kernel")
            kernel = self.kernel = self.compute_kernel()
            return kernel

        elif name == "precond_kernel":
            if self.preconditioner == "circulant":
                logger.info("Computing Preconditioner kernel")
                precond_kernel = self.precond_kernel = FourierKernel(
                    1.0 / self.kernel.circularize(), centered=True
                )
            else:
                precond_kernel = self.precond_kernel = None
            return precond_kernel

        return super(Estimator, self).__getattr__(name)

    def compute_kernel(self):
        raise NotImplementedError("Subclasses must implement the compute_kernel method")

    def estimate(self, b_coeff=None, tol=1e-5, regularizer=0):
        """Return an estimate as a Volume instance."""
        if b_coeff is None:
            b_coeff = self.src_backward()
        # conj_grad expects a 1d array if n = 1
        b_coeff = np.squeeze(b_coeff, axis=0)
        est_coeff = self.conj_grad(b_coeff, tol=tol, regularizer=regularizer)
        est = self.basis.evaluate(est_coeff).T

        return est

    def src_backward(self):
        """
        Apply adjoint mapping to source

        :return: The adjoint mapping applied to the images, averaged over the whole dataset and expressed
            as coefficients of `basis`.
        """
        mean_b = Volume(
            np.zeros((self.src.L, self.src.L, self.src.L), dtype=self.dtype)
        )

        for i in range(0, self.src.n, self.batch_size):
            im = self.src.images[i : i + self.batch_size]
            batch_mean_b = self.src.im_backward(im, i) / self.src.n
            mean_b += batch_mean_b.astype(self.dtype)

        res = self.basis.evaluate_t(mean_b)
        logger.info(f"Determined adjoint mappings. Shape = {res.shape}")
        return res

    def conj_grad(self, b_coeff, tol=1e-5, regularizer=0):
        n = b_coeff.shape[0]
        kernel = self.kernel

        if regularizer > 0:
            kernel += regularizer

        operator = LinearOperator(
            (n, n), matvec=partial(self.apply_kernel, kernel=kernel), dtype=self.dtype
        )
        if self.precond_kernel is None:
            M = None
        else:
            precond_kernel = self.precond_kernel
            if regularizer > 0:
                precond_kernel += regularizer
            M = LinearOperator(
                (n, n),
                matvec=partial(self.apply_kernel, kernel=precond_kernel),
                dtype=self.dtype,
            )

        target_residual = tol * norm(b_coeff)

        def cb(xk):
            logger.info(
                f"Delta {norm(b_coeff - self.apply_kernel(xk))} (target {target_residual})"
            )

        x, info = scipy.sparse.linalg.cg(
            operator, b_coeff, M=M, callback=cb, tol=tol, atol=0
        )

        if info != 0:
            raise RuntimeError("Unable to converge!")
        return x

    def apply_kernel(self, vol_coeff, kernel=None):
        """
        Applies the kernel represented by convolution

        :param vol_coeff: The volume to be convolved, stored in the basis coefficients.
        :param kernel: a Kernel object. If None, the kernel for this Estimator is used.
        :return: The result of evaluating `vol_coeff` in the given basis, convolving with the kernel given by
            kernel, and backprojecting into the basis.
        """
        if kernel is None:
            kernel = self.kernel
        vol = self.basis.evaluate(vol_coeff)
        # convolve_volume expects a 3-dimensional array
        # so we remove the first dimension of the volume, which is 1
        vol = Volume(kernel.convolve_volume(vol[0]))
        vol = self.basis.evaluate_t(vol)

        return vol



"""
Utilities for controlling and generating random numbers.
"""

import numpy as np
from scipy.special import erfinv

from aspire.utils.matlab_compat import m_reshape

# A list of random states, used as a stack
random_states = []


def choice(*args, **kwargs):
    """
    Wraps numpy random.choice call in ASPIRE Random context.
    """
    seed = None
    if "seed" in kwargs:
        seed = kwargs.pop("seed")

    with Random(seed):
        return np.random.choice(*args, **kwargs)


def randi(i_max, size, seed=None):
    """
    A MATLAB compatible randi implementation that returns numbers from a discrete uniform distribution.
    While a direct use of np.random.choice would be convenient, this doesn't seem to return results
    identical to MATLAB.

    :param iMax: TODO
    :param size: size of the resulting np array
    :param seed: Random seed to use (None to apply no seed)
    :return: A np array
    """
    with Random(seed):
        return np.ceil(i_max * np.random.random(size=size)).astype("int")


def randn(*args, **kwargs):
    """
    Calls rand and applies inverse transform sampling to the output.
    """
    seed = None
    if "seed" in kwargs:
        seed = kwargs.pop("seed")

    with Random(seed):
        uniform = np.random.rand(*args, **kwargs)
        result = np.sqrt(2) * erfinv(2 * uniform - 1)
        # TODO: Rearranging elements to get consistent behavior with MATLAB 'randn2'
        result = m_reshape(result.flatten(), args)
        return result


def rand(size, seed=None):
    """
    Note this is for MATLAB repro (see m_reshape).

    Other uses prefer use of `random`.
    """
    with Random(seed):
        return m_reshape(np.random.random(np.prod(size)), size)


def random(*args, **kwargs):
    """
    Wraps numpy.random.random with ASPIRE Random context manager.
    """
    seed = kwargs.pop("seed", None)

    with Random(seed):
        return np.random.random(*args, **kwargs)


class Random:
    """
    A context manager that pushes a random seed to the stack for reproducible results,
    and pops it on exit.
    """

    def __init__(self, seed=None):
        self.seed = seed

    def __enter__(self):
        if self.seed is not None:
            # Push current state on stack
            random_states.append(np.random.get_state())

            seed = self.seed
            # 5489 is the default seed used by MATLAB for seed 0 !
            if seed == 0:
                seed = 5489

            new_state = np.random.RandomState(seed)
            np.random.set_state(new_state.get_state())

    def __exit__(self, *args):
        if self.seed is not None:
            np.random.set_state(random_states.pop())



import json
import logging
import os

import click

from aspire.apple.apple import Apple
from aspire.commands import log_level_option
from aspire.utils.logging import setConsoleLoggingLevel

logger = logging.getLogger(__name__)


@click.command()
@click.option(
    "--mrc_path",
    help="Path to an mrc file or folder containing all mrcs for particle picking.",
)
@click.option("--create_jpg", is_flag=True, help="save JPG files for picked particles.")
@click.option(
    "--output_dir",
    help="Path to folder to save STAR files. If unspecified, no STAR files are created.",
    default=None,
    type=str,
)
@click.option(
    "--particle_size",
    help="Particle size in pixels.  Many other options have internal defaults based off of particle size.",
    required=True,
    type=int,
)
@click.option("--max_particle_size", help="", default=None, type=int)
@click.option("--min_particle_size", help="", default=None, type=int)
@click.option("--minimum_overlap_amount", help="", default=None, type=int)
@click.option("--query_image_size", help="", default=None)
@click.option("--tau1", help="", default=None, type=int)
@click.option("--tau2", help="", default=None, type=int)
@click.option("--container_size", help="", default=450)
@click.option("--model", help="", default="svm")
@click.option(
    "--model_opts",
    help="Optional JSON dictionary representing specific options corresponding to `model`.",
    default=None,
    type=str,
)
@click.option("--mrc_margin_left", help="", default=99)
@click.option("--mrc_margin_right", help="", default=100)
@click.option("--mrc_margin_top", help="", default=99)
@click.option("--mrc_margin_bottom", help="", default=100)
@click.option("--mrc_shrink_factor", help="", default=2)
@click.option("--mrc_gauss_filter_size", help="", default=15)
@click.option("--mrc_gauss_filter_sigma", help="", default=0.5)
@click.option("--response_thresh_norm_factor", help="", default=20)
@click.option("--conv_map_nthreads", help="", default=4)
@click.option(
    "--n_processes",
    help="Concurrent processes to spawn."
    "May improve performance on very large machines."
    "Otherwise use default.",
    default=1,
)
@log_level_option
def apple(
    mrc_path,
    create_jpg,
    output_dir,
    particle_size,
    max_particle_size,
    min_particle_size,
    query_image_size,
    tau1,
    tau2,
    minimum_overlap_amount,
    container_size,
    model,
    model_opts,
    mrc_margin_left,
    mrc_margin_right,
    mrc_margin_top,
    mrc_margin_bottom,
    mrc_shrink_factor,
    mrc_gauss_filter_size,
    mrc_gauss_filter_sigma,
    response_thresh_norm_factor,
    conv_map_nthreads,
    n_processes,
    loglevel,
):
    """Pick and save particles from one or more mrc files."""

    # Set desired logging option for the command line
    setConsoleLoggingLevel(loglevel)

    # Convert model_opts string to a dictionary
    if model_opts is not None:
        try:
            model_opts = json.loads(model_opts)
        except Exception as e:
            logger.error(
                f"Failed to parse `model_opts`={model_opts}",
                "  Ensure well formed JSON.",
            )
            raise e

    picker = Apple(
        particle_size,
        output_dir,
        min_particle_size,
        max_particle_size,
        query_image_size,
        minimum_overlap_amount,
        tau1,
        tau2,
        container_size,
        model,
        model_opts,
        mrc_margin_left,
        mrc_margin_right,
        mrc_margin_top,
        mrc_margin_bottom,
        mrc_shrink_factor,
        mrc_gauss_filter_size,
        mrc_gauss_filter_sigma,
        response_thresh_norm_factor,
        conv_map_nthreads,
        n_processes,
    )

    if not os.path.exists(mrc_path):
        raise RuntimeError(f"`mrc_path` does not exist: {mrc_path}")
    elif os.path.isdir(mrc_path):
        func = picker.process_folder
    else:
        func = picker.process_micrograph

    func(mrc_path, create_jpg=create_jpg)



import logging

from aspire.abinitio import CLOrient3D

logger = logging.getLogger(__name__)


class CommLineLUD(CLOrient3D):
    """
    Define a derived class to estimate 3D orientations using Least Unsquared Deviations described as below:
    L. Wang, A. Singer, and  Z. Wen, Orientation Determination of Cryo-EM Images Using Least Unsquared Deviations,
    SIAM J. Imaging Sciences, 6, 2450-2483 (2013).

    """

    def __init__(self, src):
        """
        constructor of an object for estimating 3D orientations
        """
        pass

    def estimate(self):
        """
        perform estimation of orientations
        """
        pass

    def output(self):
        """
        Output the 3D orientations
        """
        pass



import logging

import click

from aspire.commands import log_level_option
from aspire.noise import WhiteNoiseEstimator
from aspire.source.relion import RelionSource
from aspire.utils.logging import setConsoleLoggingLevel

logger = logging.getLogger(__name__)


@click.command()
@click.option("--data_folder", default=None, help="Path to data folder")
@click.option(
    "--starfile_in",
    required=True,
    help="Path to input STAR file relative to data folder",
)
@click.option(
    "--starfile_out",
    required=True,
    help="Path to output STAR file relative to data folder",
)
@click.option(
    "--pixel_size", default=1, type=float, help="Pixel size of images in STAR file"
)
@click.option(
    "--max_rows",
    default=None,
    type=int,
    help="Max number of image rows to read from STAR file",
)
@click.option("--flip_phase", default=True, help="Perform phase flip or not")
@click.option(
    "--downsample",
    default=None,
    type=int,
    help="Downsample the images to this resolution prior to saving to starfile/.mrcs stack",
)
@click.option(
    "--normalize_bg",
    default=True,
    help="Normalize the images to have mean zero and variance one in the corners",
)
@click.option(
    "--whiten",
    default=True,
    help="Estimate the noise variance of the images and whiten",
)
@click.option(
    "--invert_contrast",
    default=True,
    help="Invert the contrast of the images to ensure that clean particles have positive intensity",
)
@click.option(
    "--batch_size", default=512, help="Batch size to load images from MRC files"
)
@click.option(
    "--save_mode",
    default="single",
    help="Option to save MRC file, if not single, saved to multiple files in batch size",
)
@click.option(
    "--overwrite",
    default=False,
    help="Whether to overwrite MRC files if they already exist",
)
@log_level_option
def preprocess(
    data_folder,
    starfile_in,
    starfile_out,
    pixel_size,
    max_rows,
    flip_phase,
    downsample,
    normalize_bg,
    whiten,
    invert_contrast,
    batch_size,
    save_mode,
    overwrite,
    loglevel,
):
    """
    Preprocess the raw images and output desired images for future analysis
    """
    # Set desired logging option for the command line
    setConsoleLoggingLevel(loglevel)

    # Create a source object for 2D images
    logger.info(f"Read in images from {starfile_in} and preprocess the images.")
    source = RelionSource(
        starfile_in, data_folder, pixel_size=pixel_size, max_rows=max_rows
    )

    if flip_phase:
        logger.info("Perform phase flip to input images")
        source = source.phase_flip()

    if downsample and downsample < source.L:
        logger.info(f"Downsample resolution to {downsample} X {downsample}")
        source = source.downsample(downsample)

    if normalize_bg:
        logger.info("Normalize images to noise background")
        source = source.normalize_background()

    if whiten:
        logger.info("Whiten noise of images")
        noise_estimator = WhiteNoiseEstimator(source)
        source = source.whiten(noise_estimator)

    if invert_contrast:
        logger.info("Invert global density contrast")
        source = source.invert_contrast()

    source.save(
        starfile_out, batch_size=batch_size, save_mode=save_mode, overwrite=overwrite
    )



import logging

import numpy as np
from numpy.linalg import solve

from aspire.basis import FFBBasis2D
from aspire.covariance import BatchedRotCov2D
from aspire.denoising import Denoiser
from aspire.denoising.denoised_src import DenoisedImageSource
from aspire.noise import WhiteNoiseEstimator
from aspire.optimization import fill_struct
from aspire.utils import mat_to_vec
from aspire.volume import Volume, qr_vols_forward

logger = logging.getLogger(__name__)


def src_wiener_coords(
    sim, mean_vol, eig_vols, lambdas=None, noise_var=0, batch_size=512
):
    """
    Calculate coordinates using Wiener filter

    :param sim: A simulation object containing the images whose coordinates we want.
    :param mean_vol: The mean volume of the source as Volume instance.
    :param eig_vols: The eigenvolumes of the source as Volume instance.
    :param lambdas: The eigenvalues in a K-by-K diagonal matrix (default `eye(K)`).
    :param noise_var: The variance of the noise in the images (default 0).
    :param batch_size: The size of the batches in which to compute the coordinates (default 512).
    :return: A K-by-`src.n` array of coordinates corresponding to the Wiener filter coordinates of each image in sim.

    The coordinates are obtained by the formula
        alpha_s = eig_vols^T H_s ( y_s - P_s mean_vol ) ,

    where P_s is the forward image mapping and y_s is the sth image,
        H_s = Sigma * P_s^T ( P_s Sigma P_s^T + noise_var I )^(-1) ,

    and Sigma is the covariance matrix eig_vols * lambdas * eig_vols^T.
    Note that when noise_var is zero, this reduces to the projecting y_s onto the span of P_s eig_vols.

    # TODO: Find a better place for this functionality other than in utils
    """

    if not isinstance(mean_vol, Volume):
        logger.debug(
            "src_wiener_coords mean_vol should be a Volume instance. Attempt correction."
        )
        if len(mean_vol.shape) == 4 and mean_vol.shape[3] != 1:
            msg = (
                f"Cannot naively convert {mean_vol.shape} to Volume instance."
                "Please change calling code."
            )
            logger.error(msg)
            raise RuntimeError(msg)

        mean_vol = Volume(mean_vol)

    if not isinstance(eig_vols, Volume):
        logger.debug(
            "src_wiener_coords eig_vols should be a Volume instance. Correcting for now."
        )
        eig_vols = Volume(eig_vols)

    if not sim.dtype == mean_vol.dtype == eig_vols.dtype:
        logger.warning(
            "Inconsistent types in src_wiener_coords"
            f" sim {sim.dtype},"
            f" mean_vol {mean_vol.dtype},"
            f" eig_vols {eig_vols.dtype}"
        )

    k = eig_vols.n_vols
    if lambdas is None:
        lambdas = np.eye(k)

    coords = np.zeros((k, sim.n))
    covar_noise = noise_var * np.eye(k)

    for i in range(0, sim.n, batch_size):
        ims = sim.images[i : i + batch_size]
        batch_n = ims.shape[0]
        ims -= sim.vol_forward(mean_vol, i, batch_n)

        Qs, Rs = qr_vols_forward(sim, i, batch_n, eig_vols, k)

        Q_vecs = mat_to_vec(Qs)

        # RCOPT
        ims = np.moveaxis(ims.asnumpy(), 0, 2)
        im_vecs = mat_to_vec(ims)

        for j in range(batch_n):
            im_coords = Q_vecs[:, :, j].T @ im_vecs[:, j]
            covar_im = (Rs[:, :, j] @ lambdas @ Rs[:, :, j].T) + covar_noise
            xx = solve(covar_im, im_coords)
            im_coords = lambdas @ Rs[:, :, j].T @ xx
            coords[:, i + j] = im_coords

    return coords


class DenoiserCov2D(Denoiser):
    """
    Define a derived class for denoising 2D images using Cov2D method
    """

    def __init__(self, src, basis=None, var_noise=None):
        """
        Initialize an object for denoising 2D images using Cov2D method

        :param src: The source object of 2D images with metadata
        :param basis: The basis method to expand 2D images
        :param var_noise: The estimated variance of noise
        """
        super().__init__(src)

        # When var_noise is not specfically over-ridden,
        #   recompute it now. See #496.
        if var_noise is None:
            logger.info("Estimating noise of images using WhiteNoiseEstimator")
            noise_estimator = WhiteNoiseEstimator(src)
            var_noise = noise_estimator.estimate()
            logger.info(f"Estimated Noise Variance: {var_noise}")
        self.var_noise = var_noise

        if basis is None:
            basis = FFBBasis2D((self.src.L, self.src.L), dtype=src.dtype)

        if not isinstance(basis, FFBBasis2D):
            raise NotImplementedError("Currently only fast FB method is supported")

        self.basis = basis
        self.cov2d = None
        self.mean_est = None
        self.covar_est = None

    def denoise(self, covar_opt=None, batch_size=512):
        """
         Build covariance matrix of 2D images and return a new ImageSource object

        :param covar_opt: The option list for building Cov2D matrix
        :param batch_size: The batch size for processing images
        :return: A `DenoisedImageSource` object with the specified denoising object
        """

        # Initialize the rotationally invariant covariance matrix of 2D images
        # A fixed batch size is used to go through each image
        self.cov2d = BatchedRotCov2D(self.src, self.basis, batch_size=batch_size)

        default_opt = {
            "shrinker": "frobenius_norm",
            "verbose": 0,
            "max_iter": 250,
            "iter_callback": [],
            "store_iterates": False,
            "rel_tolerance": 1e-12,
            "precision": self.dtype,
        }

        covar_opt = fill_struct(covar_opt, default_opt)
        # Calculate the mean and covariance for the rotationally invariant covariance matrix of 2D images
        self.mean_est = self.cov2d.get_mean()

        self.covar_est = self.cov2d.get_covar(
            noise_var=self.var_noise, mean_coeff=self.mean_est, covar_est_opt=covar_opt
        )

        return DenoisedImageSource(self.src, self)

    def images(self, istart=0, batch_size=512):
        """
        Obtain a batch size of 2D images after denosing by Cov2D method

        :param istart: the index of starting image
        :param batch_size: The batch size for processing images
        :return: an `Image` object with denoised images
        """
        src = self.src

        # Denoise one batch size of 2D images using the SPCAs from the rotationally invariant covariance matrix
        img_start = istart
        img_end = min(istart + batch_size, src.n)
        imgs_noise = src.images[img_start : img_start + batch_size]
        coeffs_noise = self.basis.evaluate_t(imgs_noise)
        logger.info(
            f"Estimating Cov2D coefficients for images from {img_start} to {img_end-1}"
        )
        coeffs_estim = self.cov2d.get_cwf_coeffs(
            coeffs_noise,
            self.cov2d.ctf_fb,
            self.cov2d.ctf_idx[img_start:img_end],
            mean_coeff=self.mean_est,
            covar_coeff=self.covar_est,
            noise_var=self.var_noise,
        )

        # Convert Fourier-Bessel coefficients back into 2D images
        logger.info("Converting Cov2D coefficients back to 2D images")
        imgs_denoised = self.basis.evaluate(coeffs_estim)

        return imgs_denoised



import logging

import mrcfile
import numpy as np
from numpy.linalg import qr

import aspire.image
from aspire.nufft import nufft
from aspire.numeric import fft, xp
from aspire.utils import Rotation, crop_pad_3d, grid_2d, grid_3d, mat_to_vec, vec_to_mat
from aspire.utils.types import complex_type

logger = logging.getLogger(__name__)


def qr_vols_forward(sim, s, n, vols, k):
    """
    TODO: Write docstring
    TODO: Find a better place for this!

    :param sim:
    :param s:
    :param n:
    :param vols:
    :param k:
    :return:
    """
    ims = np.zeros((k, n, sim.L, sim.L), dtype=vols.dtype)
    for ell in range(k):
        ims[ell] = sim.vol_forward(vols[ell], s, n).asnumpy()

    ims = np.swapaxes(ims, 1, 3)
    ims = np.swapaxes(ims, 0, 2)

    Q_vecs = np.zeros((sim.L**2, k, n), dtype=vols.dtype)
    Rs = np.zeros((k, k, n), dtype=vols.dtype)

    im_vecs = mat_to_vec(ims)
    for i in range(n):
        Q_vecs[:, :, i], Rs[:, :, i] = qr(im_vecs[:, :, i])
    Qs = vec_to_mat(Q_vecs)

    return Qs, Rs


class Volume:
    """
    Volume is an (N1 x ...) x L x L x L array, along with associated utility methods.
    """

    def __init__(self, data, dtype=None):
        """
        A stack of one or more volumes.

        This is a wrapper of numpy.ndarray which provides methods
        for common processing tasks.

        The stack can be multidimensional with `n_vols` equal
        to the product of the stack dimensions.  Singletons will be
        expanded into a stack with one entry.

        The last three axes represent the volume size,
        and are checked to be cubic.

        :param data: Numpy array containing volume data with shape
            `(..., resolution, resolution, resolution)`.
        :param dtype: Optionally cast `data` to this dtype.
            Defaults to `data.dtype`.

        :return: A Volume instance holding `data`.
        """

        if not isinstance(data, np.ndarray):
            raise ValueError("Volume should be instantiated with an ndarray")

        if data.ndim < 3:
            raise ValueError(
                "Volume data should be ndarray with shape (N1...)xLxLxL or LxLxL."
            )
        elif data.ndim == 3:
            data = np.expand_dims(data, axis=0)

        if dtype is None:
            self.dtype = data.dtype
        else:
            self.dtype = np.dtype(dtype)

        if not (data.shape[-1] == data.shape[-2] == data.shape[-3]):
            raise ValueError("Only cubed ndarrays are supported.")

        self._data = data.astype(self.dtype, copy=False)
        self.ndim = self._data.ndim
        self.shape = self._data.shape
        self.stack_ndim = self._data.ndim - 3
        self.stack_shape = self._data.shape[:-3]
        self.n_vols = np.prod(self.stack_shape)
        self.resolution = self._data.shape[-1]

        # Numpy interop
        # https://numpy.org/devdocs/user/basics.interoperability.html#the-array-interface-protocol
        self.__array_interface__ = self._data.__array_interface__
        self.__array__ = self._data

    def asnumpy(self):
        """
        Return volume data as a (<stack>, resolution, resolution,
        resolution) read-only array view.

        :return: read-only ndarray view
        """
        view = self._data.view()
        view.flags.writeable = False
        return view

    def astype(self, dtype, copy=True):
        """
        Return `Volume` instance with the prescribed dtype.

        :param dtype: Numpy dtype
        :param copy: Boolean, optionally avoid copying if Volume.dtype already matches.
            Defaults to True.
        :return: Volume instance
        """
        return self.__class__(self.asnumpy().astype(dtype, copy=copy))

    def _check_key_dims(self, key):
        if isinstance(key, tuple) and (len(key) > self._data.ndim):
            raise ValueError(
                f"Volume stack_dim is {self.stack_ndim}, slice length must be =< {self.ndim}"
            )

    def __getitem__(self, key):
        self._check_key_dims(key)
        return self.__class__(self._data[key])

    def __setitem__(self, key, value):
        self._check_key_dims(key)
        self._data[key] = value

    def stack_reshape(self, *args):
        """
        Reshape the stack axis.

        :*args: Integer(s) or tuple describing the intended shape.

        :returns: Volume instance
        """

        # If we're passed a tuple, use that
        if len(args) == 1 and isinstance(args[0], tuple):
            shape = args[0]
        else:
            # Otherwise use the variadic args
            shape = args

        # Sanity check the size
        if shape != (-1,) and np.prod(shape) != self.n_vols:
            raise ValueError(
                f"Number of volumes {self.n_vols} cannot be reshaped to {shape}."
            )

        return self.__class__(self._data.reshape(*shape, *self._data.shape[-3:]))

    def __repr__(self):
        msg = (
            f"{self.n_vols} {self.dtype} volumes arranged as a {self.stack_shape} stack"
        )
        msg += f" each of size {self.resolution}x{self.resolution}x{self.resolution}."
        return msg

    def __len__(self):
        return self.n_vols

    def __add__(self, other):
        if isinstance(other, Volume):
            res = self.__class__(self._data + other.asnumpy())
        else:
            res = self.__class__(self._data + other)

        return res

    def __radd__(self, otherL):
        return self + otherL

    def __sub__(self, other):
        if isinstance(other, Volume):
            res = self.__class__(self._data - other.asnumpy())
        else:
            res = self.__class__(self._data - other)

        return res

    def __rsub__(self, otherL):
        return self.__class__(otherL - self._data)

    def __mul__(self, other):
        if isinstance(other, Volume):
            res = self.__class__(self._data * other.asnumpy())
        else:
            res = self.__class__(self._data * other)

        return res

    def __rmul__(self, otherL):
        return self * otherL

    def __truediv__(self, other):
        """
        Scalar division, follows numpy semantics.
        """
        if isinstance(other, Volume):
            res = self.__class__(self._data / other.asnumpy())
        else:
            res = self.__class__(self._data / other)

        return res

    def __rtruediv__(self, otherL):
        """
        Right scalar division, follows numpy semantics.
        """
        return otherL * Volume(1.0 / self._data)

    def project(self, vol_idx, rot_matrices):
        """
        Using the stack of rot_matrices,
        project images of Volume[vol_idx].

        :param vol_idx: Volume index
        :param rot_matrices: Stack of rotations. Rotation or ndarray instance.
        :return: `Image` instance.
        """
        # See Issue #727
        if self.stack_ndim > 1:
            raise NotImplementedError(
                "`project` is currently limited to 1D Volume stacks."
            )

        # If we are an ASPIRE Rotation, get the numpy representation.
        if isinstance(rot_matrices, Rotation):
            rot_matrices = rot_matrices.matrices

        if rot_matrices.dtype != self.dtype:
            logger.warning(
                f"{self.__class__.__name__}"
                f" rot_matrices.dtype {rot_matrices.dtype}"
                f" != self.dtype {self.dtype}."
                " In the future this will raise an error."
            )

        data = self[vol_idx].asnumpy()

        n = rot_matrices.shape[0]

        pts_rot = rotated_grids(self.resolution, rot_matrices)

        # TODO: rotated_grids might as well give us correctly shaped array in the first place
        pts_rot = pts_rot.reshape((3, n * self.resolution**2))

        im_f = nufft(data, pts_rot) / self.resolution

        im_f = im_f.reshape(-1, self.resolution, self.resolution)

        if self.resolution % 2 == 0:
            im_f[:, 0, :] = 0
            im_f[:, :, 0] = 0

        im_f = xp.asnumpy(fft.centered_ifft2(xp.asarray(im_f)))

        return aspire.image.Image(np.real(im_f))

    def to_vec(self):
        """Returns an N x resolution ** 3 array."""
        return self._data.reshape((self.n_vols, self.resolution**3))

    @classmethod
    def from_vec(cls, vec):
        """
        Returns a Volume instance from a (N, resolution**3) array or
        (resolution**3) array.

        :return: Volume instance.
        """

        if vec.ndim == 1:
            vec = vec[np.newaxis, :]

        n_vols = vec.shape[0]

        resolution = round(vec.shape[1] ** (1 / 3))
        assert resolution**3 == vec.shape[1]

        data = vec.reshape((n_vols, resolution, resolution, resolution))

        return cls(data)

    def transpose(self):
        """
        Returns a new Volume instance with volume data axes tranposed.

        :return: Volume instance.
        """
        original_stack_shape = self.stack_shape
        v = self.stack_reshape(-1)
        vt = np.transpose(v._data, (0, -1, -2, -3))
        return self.__class__(vt).stack_reshape(original_stack_shape)

    @property
    def T(self):
        """
        Abbreviation for transpose.

        :return: Volume instance.
        """

        return self.transpose()

    def flatten(self):
        """
        Util function for flatten operation on volume data array.

        :return: ndarray
        """

        return self._data.flatten()

    def flip(self, axis=-3):
        """
        Flip volume stack data along axis using numpy.flip

        :param axis: Optionally specify axis as integer or tuple.
            Defaults to axis=-3.

        :return: Volume instance.
        """
        # Convert integer to tuple, so we can always loop.
        if isinstance(axis, int):
            axis = (axis,)

        for ax in axis:
            ax = ax % self.ndim  # modulo [0, ndim)
            if ax < self.stack_ndim:
                raise ValueError(
                    f"Cannot flip axis {ax}: stack axis. Did you mean {ax-4}?"
                )

        return self.__class__(np.flip(self._data, axis))

    def downsample(self, ds_res, mask=None):
        """
        Downsample each volume to a desired resolution (only cubic supported).

        :param ds_res: Desired resolution.
        :param mask: Optional NumPy array mask to multiply in Fourier space.
        """
        if mask is None:
            mask = 1.0

        original_stack_shape = self.stack_shape
        v = self.stack_reshape(-1)

        # take 3D Fourier transform of each volume in the stack
        fx = fft.fftshift(fft.fftn(v._data, axes=(1, 2, 3)))
        # crop each volume to the desired resolution in frequency space
        crop_fx = (
            np.array([crop_pad_3d(fx[i, :, :, :], ds_res) for i in range(self.n_vols)])
            * mask
        )
        # inverse Fourier transform of each volume
        out = fft.ifftn(fft.ifftshift(crop_fx), axes=(1, 2, 3)) * (
            ds_res**3 / self.resolution**3
        )
        # returns a new Volume object
        return self.__class__(np.real(out)).stack_reshape(original_stack_shape)

    def shift(self):
        raise NotImplementedError

    def rotate(self, rot_matrices, zero_nyquist=True):
        """
        Rotate volumes using a `Rotation` object. If the `Rotation` object
        is a single rotation, each volume will be rotated by that rotation.
        If the `Rotation` object is a stack of rotations of length n_vols,
        the ith volume is rotated by the ith rotation.

        :param rot_matrices: `Rotation` object of length 1 or n_vols.
        :param zero_nyquist: Option to keep or remove Nyquist frequency for even resolution.
            Defaults to zero_nyquist=True, removing the Nyquist frequency.

        :return: `Volume` instance.
        """
        if self.stack_ndim > 1:
            raise NotImplementedError(
                "`rotation` is currently limited to 1D Volume stacks."
            )

        assert isinstance(
            rot_matrices, Rotation
        ), f"Argument must be an instance of the Rotation class. {type(rot_matrices)} was supplied."

        # Get numpy representation of Rotation object.
        rot_matrices = rot_matrices.matrices

        K = len(rot_matrices)  # Rotation stack size
        assert K == self.n_vols or K == 1, "Rotation object must be length 1 or n_vols."

        if rot_matrices.dtype != self.dtype:
            logger.warning(
                f"{self.__class__.__name__}"
                f" rot_matrices.dtype {rot_matrices.dtype}"
                f" != self.dtype {self.dtype}."
                " In the future this will raise an error."
            )

        # If K = 1 we broadcast the single Rotation object across each volume.
        if K == 1:
            pts_rot = rotated_grids_3d(self.resolution, rot_matrices)
            vol_f = nufft(self.asnumpy(), pts_rot)
            vol_f = vol_f.reshape(-1, self.resolution, self.resolution, self.resolution)

        # If K = n_vols, we apply the ith rotation to ith volume.
        else:
            rot_matrices = rot_matrices.reshape((K, 1, 3, 3))
            pts_rot = np.zeros((K, 3, self.resolution**3), dtype=self.dtype)
            vol_f = np.empty(
                (self.n_vols, self.resolution**3), dtype=complex_type(self.dtype)
            )
            for i in range(K):
                pts_rot[i] = rotated_grids_3d(self.resolution, rot_matrices[i])

                vol_f[i] = nufft(self[i].asnumpy(), pts_rot[i])

            vol_f = vol_f.reshape(-1, self.resolution, self.resolution, self.resolution)

        # If resolution is even, we zero out the nyquist frequency by default.
        if self.resolution % 2 == 0 and zero_nyquist is True:
            vol_f[:, 0, :, :] = 0
            vol_f[:, :, 0, :] = 0
            vol_f[:, :, :, 0] = 0

        vol = xp.asnumpy(
            np.real(fft.centered_ifftn(xp.asarray(vol_f), axes=(-3, -2, -1)))
        )

        return self.__class__(vol)

    def denoise(self):
        raise NotImplementedError

    def save(self, filename, overwrite=False):
        """
        Save volume to disk as mrc file

        :param filename: Filepath where volume will be saved

        :param overwrite: Option to overwrite file when set to True.
            Defaults to overwrite=False.
        """
        if self.stack_ndim > 1:
            raise NotImplementedError(
                "`save` is currently limited to 1D Volume stacks."
            )

        with mrcfile.new(filename, overwrite=overwrite) as mrc:
            mrc.set_data(self._data.astype(np.float32))

        if self.dtype != np.float32:
            logger.info(f"Volume with dtype {self.dtype} saved with dtype float32")

    @classmethod
    def load(cls, filename, permissive=True, dtype=np.float32):
        """
        Load an mrc file as a Volume instance.

        :param filename: Data filepath to load.
        :param permissive: Allows problematic files to load with warning when True.
            Defaults to permissive=True.
        :param dtype: Optionally specifiy data type. Defaults to dtype=np.float32.

        :return: Volume instance.
        """
        with mrcfile.open(filename, permissive=permissive) as mrc:
            loaded_data = mrc.data
        if loaded_data.dtype != dtype:
            logger.info(f"{filename} with dtype {loaded_data.dtype} loaded as {dtype}")
        return cls(loaded_data.astype(dtype))


class CartesianVolume(Volume):
    def expand(self, basis):
        return BasisVolume(basis)


class PolarVolume(Volume):
    def expand(self, basis):
        return BasisVolume(basis)


class BispecVolume(Volume):
    def expand(self, basis):
        return BasisVolume(basis)


class BasisVolume(Volume):
    def __init__(self, basis):
        self.basis = basis

    def evaluate(self):
        return CartesianVolume()


class FBBasisVolume(BasisVolume):
    pass


# TODO: The following functions likely all need to be moved inside the Volume class


def rotated_grids(L, rot_matrices):
    """
    Generate rotated Fourier grids in 3D from rotation matrices

    :param L: The resolution of the desired grids.
    :param rot_matrices: An array of size k-by-3-by-3 containing K rotation matrices
    :return: A set of rotated Fourier grids in three dimensions as specified by the rotation matrices.
        Frequencies are in the range [-pi, pi].
    """

    grid2d = grid_2d(L, indexing="xy", dtype=rot_matrices.dtype)
    num_pts = L**2
    num_rots = rot_matrices.shape[0]
    pts = np.pi * np.vstack(
        [
            grid2d["x"].flatten(),
            grid2d["y"].flatten(),
            np.zeros(num_pts, dtype=rot_matrices.dtype),
        ]
    )
    pts_rot = np.zeros((3, num_rots, num_pts), dtype=rot_matrices.dtype)
    for i in range(num_rots):
        pts_rot[:, i, :] = rot_matrices[i, :, :] @ pts

    pts_rot = pts_rot.reshape((3, num_rots, L, L))

    return pts_rot


def rotated_grids_3d(L, rot_matrices):
    """
    Generate rotated Fourier grids in 3D from rotation matrices.

    :param L: The resolution of the desired grids.
    :param rot_matrices: An array of size k-by-3-by-3 containing K rotation matrices
    :return: A set of rotated Fourier grids in three dimensions as specified by the rotation matrices.
        Frequencies are in the range [-pi, pi].
    """

    grid3d = grid_3d(L, indexing="xyz", dtype=rot_matrices.dtype)
    num_pts = L**3
    num_rots = rot_matrices.shape[0]
    pts = np.pi * np.vstack(
        [
            grid3d["x"].flatten(),
            grid3d["y"].flatten(),
            grid3d["z"].flatten(),
        ]
    )
    pts_rot = np.zeros((3, num_rots, num_pts), dtype=rot_matrices.dtype)
    for i in range(num_rots):
        pts_rot[:, i, :] = rot_matrices[i, :, :] @ pts

    # Note we return grids as (Z,Y,X)
    return pts_rot.reshape(3, -1)



"""
Miscellaneous Utilities that have no better place (yet).
"""
import hashlib
import importlib.resources
import logging
import sys
from itertools import chain, combinations

import numpy as np
from scipy.special import erf

from aspire.utils import grid_1d, grid_2d, grid_3d
from aspire.utils.rotation import Rotation

logger = logging.getLogger(__name__)


def importlib_path(package, resource):
    """
    Return the path to the resource as an actual file system path.
    Workaround importlib.resources deprecation of `path` in Python 3.11.
    This is expected to be safely removed after the minimal supported
    Python is 3.9.

    See ASPIRE-Python #546.

    :param package: Is either a name or a module object
        which conforms to the Package requirements.
    :param resource: Is the name of the resource to open within package;
        It may not contain path separators and it may not have sub-resources.
        (i.e. it cannot be a directory)
    :return: This function returns a context manager for use in a with statement.
        The context manager provides a pathlib.Path object.
    """

    py_version = sys.version_info

    # Use the deprecated method
    if py_version.major == 3 and py_version.minor < 9:
        p = importlib.resources.path(package, resource)
    else:
        p = importlib.resources.as_file(
            importlib.resources.files(package).joinpath(resource)
        )

    return p


def abs2(x):
    """
    Compute complex modulus squared.
    """

    return x.real**2 + x.imag**2


def powerset(iterable):
    """
    Generate all subsets of an iterable. Example:

    powerset([1,2,3]) --> () (1,) (2,) (3,) (1,2) (1,3) (2,3) (1,2,3)

    :return: Generator covering all subsets of iterable.
    """

    s = list(iterable)
    return chain.from_iterable(combinations(s, r) for r in range(len(s) + 1))


def sha256sum(filename):
    """
    Return sha256 hash of filename.

    :param filename: path to file
    :return: sha256 hash as hex
    """

    h = hashlib.sha256()
    b = bytearray(128 * 1024)
    mv = memoryview(b)
    with open(filename, "rb", buffering=0) as f:
        for n in iter(lambda: f.readinto(mv), 0):
            h.update(mv[:n])

    return h.hexdigest()


def gaussian_1d(size, mu=0, sigma=1, dtype=np.float64):
    """
    Returns the 1D Gaussian

    .. math::
        g(x)=\\exp\\left(\\frac{ -(x - \\mu)^2}{2\\sigma^2}\\right)

    in a 1D numpy array.

    :param size: The length of the returned array (pixels)
    :param mu: mean or center (pixels)
    :param sigma: standard deviation of the Gaussian
    :param dtype: dtype of returned array
    :return: Numpy array (1D)
    """

    # Construct centered mesh
    g = grid_1d(size, normalized=False, dtype=dtype)

    p = (g["x"] - mu) ** 2 / (2 * sigma**2)

    return np.exp(-p).astype(dtype, copy=False)


def gaussian_2d(size, mu=(0, 0), sigma=(1, 1), indexing="yx", dtype=np.float64):
    """
    Returns the 2D Gaussian

    .. math::
        g(x,y)=\\exp\\left(\\frac{-(x - \\mu_x)^2}{2\\sigma_x^2} +
                \\frac{-(y - \\mu_y)^2}{2\\sigma_y^2}\\right)

    in a square 2D numpy array.

    :param size: The length of each dimension of the returned array (pixels)
    :param mu: Iterable of len(2), :math:`(\\mu_x, \\mu_y)`, indicating the center of the Gaussian
    :param sigma: Iterable of len(2) or constant, :math:`(\\sigma_x, \\sigma_y)`, of the standard
            deviation in the x and y directions. A single value, :math:`\\sigma`, can be
            used when :math:`\\sigma_x = \\sigma_y`.
    :param indexing: The order of axis indexing, passed to `aspire.utils.grid_2d`
    :param dtype: dtype of returned array
    :return: Numpy array (2D)
    """
    if np.ndim(sigma) == 0:
        sigma = (sigma, sigma)

    mu = np.array(mu, dtype=dtype)
    sigma = np.array(sigma, dtype=dtype)
    if len(mu) != 2:
        raise ValueError("`mu` must be len(2).")
    if len(sigma) != 2:
        raise ValueError("`sigma` must be a scalar or len(2).")

    # Construct centered mesh
    g = grid_2d(size, shifted=False, normalized=False, indexing=indexing, dtype=dtype)

    if indexing == "yx":
        mu, sigma = mu[::-1], sigma[::-1]
        g["x"], g["y"] = g["y"], g["x"]
    elif indexing != "xy":
        raise ValueError("Indexing must be `yx` or `xy`.")

    p = (g["x"] - mu[0]) ** 2 / (2 * sigma[0] ** 2) + (g["y"] - mu[1]) ** 2 / (
        2 * sigma[1] ** 2
    )

    return np.exp(-p).astype(dtype, copy=False)


def gaussian_3d(size, mu=(0, 0, 0), sigma=(1, 1, 1), indexing="zyx", dtype=np.float64):
    """
    Returns the 3D Gaussian

    .. math::
        g(x,y,z)=\\exp\\left(\\frac{-(x - \\mu_x)^2}{2\\sigma_x^2} +
                \\frac{-(y - \\mu_y)^2}{2\\sigma_y^2} +
                \\frac{-(z - \\mu_z)^2}{2\\sigma_z^2}\\right)

    in a 3D numpy array.

    :param size: The length of each dimension of the returned array (pixels)
    :param mu: Iterable of len(3), :math:`(\\mu_x, \\mu_y, \\mu_z)`, indicating the center of the Gaussian
    :param sigma: Iterable of len(3) or constant, :math:`(\\sigma_x, \\sigma_y, \\sigma_z)`, of the standard deviation
            in the x, y, and z directions. A single value, :math:`\\sigma`, can be
            used when :math:`\\sigma_x = \\sigma_y = \\sigma_z`
    :param indexing: The order of axis indexing, passed to `aspire.utils.grid_3d`
    :param dtype: dtype of returned array
    :return: Numpy array (3D)
    """
    if np.ndim(sigma) == 0:
        sigma = (sigma, sigma, sigma)

    mu = np.array(mu, dtype=dtype)
    sigma = np.array(sigma, dtype=dtype)
    if len(mu) != 3:
        raise ValueError("`mu` must be len(3).")
    if len(sigma) != 3:
        raise ValueError("`sigma` must be a scalar or len(3).")

    # Construct centered mesh
    g = grid_3d(size, shifted=False, normalized=False, indexing=indexing, dtype=dtype)

    if indexing == "zyx":
        mu, sigma = mu[::-1], sigma[::-1]
        g["x"], g["y"], g["z"] = g["z"], g["y"], g["x"]
    elif indexing != "xyz":
        raise ValueError("Indexing must be `zyx` or `xyz`.")

    p = (
        (g["x"] - mu[0]) ** 2 / (2 * sigma[0] ** 2)
        + (g["y"] - mu[1]) ** 2 / (2 * sigma[1] ** 2)
        + (g["z"] - mu[2]) ** 2 / (2 * sigma[2] ** 2)
    )

    return np.exp(-p).astype(dtype, copy=False)


def bump_3d(size, spread=1, dtype=np.float64):
    """
    Returns a centered 3D bump function in a (size)x(size)x(size) numpy array.
    :param size: The length of the dimensions of the array (pixels.
    :param spread: A factor controling the spread of the bump function.
    :param dtype: dtype of returned array
    :return: Numpy array (3D)
    """
    g = grid_3d(size, dtype=dtype)

    p = g["x"] ** 2 + g["y"] ** 2 + g["z"] ** 2

    bump = np.zeros((size,) * 3, dtype=dtype)
    bump[p < 1] = np.exp(-1 / (spread - spread * p[p < 1]))
    bump /= np.exp(-1 / spread)

    return bump


def circ(size, x0=0, y0=0, radius=1, peak=1, dtype=np.float64):
    """
    Returns a 2d `circ` function in a square 2d numpy array.

    where for r = sqrt(x**2 + y**2)

    circ(x,y) = peak : 0 <= r <= radius
                0 : otherwise

    Default is a centered circle of spread=peak=1.

    :param size: The height and width of returned array (pixels)
    :param x0: x cordinate of center (pixels)
    :param y0: y cordinate of center (pixels)
    :param radius: radius of circle
    :param peak: peak height at center
    :param dtype: dtype of returned array
    :return: Numpy array (2D)
    """

    # Construct centered mesh
    g = grid_2d(size, shifted=True, normalized=False, indexing="yx", dtype=dtype)

    vals = ((g["x"] - x0) ** 2 + (g["y"] - y0) ** 2) < radius * radius
    return (peak * vals).astype(dtype)


def inverse_r(size, x0=0, y0=0, peak=1, dtype=np.float64):
    """
    Returns a 2d inverse radius function in a square 2d numpy array.

    Where inverse_r(x,y): 1/sqrt(1 + x**2 + y**2)

    Default is a centered circle of peak=1.

    :param size: The height and width of returned array (pixels)
    :param x0: x cordinate of center (pixels)
    :param y0: y cordinate of center (pixels)
    :param peak: peak height at center
    :param dtype: dtype of returned array
    :return: Numpy array (2D)
    """

    # Construct centered mesh
    g = grid_2d(size, shifted=True, normalized=False, indexing="yx", dtype=dtype)

    # Compute the denominator
    vals = np.sqrt(1 + (g["x"] - x0) ** 2 + (g["y"] - y0) ** 2)

    return (peak / vals).astype(dtype)


def fuzzy_mask(L, r0, risetime, origin=None):
    """
    Create a centered 1D to 3D fuzzy mask of radius r0

    Made with an error function with effective rise time.

    :param L: The sizes of image in tuple structure
    :param r0: The specified radius
    :param risetime: The rise time for `erf` function
    :param origin: The coordinates of origin
    :return: The desired fuzzy mask
    """

    center = [sz // 2 + 1 for sz in L]
    if origin is None:
        origin = center

    grids = [np.arange(1 - org, ell - org + 1) for ell, org in zip(L, origin)]
    XYZ = np.meshgrid(*grids, indexing="ij")
    XYZ_sq = [X**2 for X in XYZ]
    R = np.sqrt(np.sum(XYZ_sq, axis=0))
    k = 1.782 / risetime
    m = 0.5 * (1 - erf(k * (R - r0)))

    return m


def all_pairs(n, return_map=False):
    """
    All pairs indexing (i,j) for i<j and a pairs-to-linear index mapping.

    :param n: The number of items to be indexed.
    :param return_map: Option to return pairs-to-linear index map.
        Default is False.
    :returns:
        - n x 2 array of pairs (i, j), i<j.
        - n x n array with pairs-to-linear index map.
    """
    pairs = np.column_stack(np.triu_indices(n, 1)).astype(np.uint16)

    if return_map:
        pairs_to_linear_map = np.empty((n, n), dtype="int")
        for index, pair in enumerate(pairs):
            pairs_to_linear_map[pair[0], pair[1]] = index
        return pairs, pairs_to_linear_map

    return pairs


def all_triplets(n):
    """
    All 3-tuples (i,j,k) where i<j<k.

    :param n: The number of items to be indexed.
    :returns: All 3-tuples (i,j,k), i<j<k.
    """
    triplets = np.fromiter(combinations(range(n), 3), dtype="int,int,int")

    return triplets


def J_conjugate(A):
    """
    Conjugate the 3x3 matrix A by the diagonal matrix J=diag((-1, -1, 1)).

    :param A: A 3x3 matrix.
    :return: J*A*J
    """
    J = np.array([[1, 1, -1], [1, 1, -1], [-1, -1, 1]], dtype=A.dtype)

    return A * J


def cyclic_rotations(order, dtype=np.float64):
    """
    Build all rotation matrices that rotate by multiples of 2pi/order about the z-axis.

    :param order: The order of cyclic symmetry
    :return: A Rotation object containing an (order)x3x3 array of rotation matrices.
    """
    angles = np.zeros((order, 3), dtype=dtype)
    angles[:, 2] = 2 * np.pi * np.arange(order) / order
    rots_symm = Rotation.from_euler(angles)

    return rots_symm


# Potentially cache this in the future.
def support_mask(L, support_radius=None, dtype=np.float64):
    """
    Return a mask selecting values within `support_radius`.

    This mask is hard cutoff, boolean type.
    For a soft cutoff, see `fuzzy_mask`

    Use for selecting signal.
    Alternatively the mask inverse (~) can be used to select background.
    Combinations can be used to create bands.

    :param L: Resolution in pixels.
    :param support_radius: Radius of mask in pixels.
        Defaults to L // 2.
    :param dtype: Dtype used for mask construction and comparison.
    :return: Boolean mask as (L,L) array.
    """

    if support_radius is None:
        support_radius = L // 2

    elif support_radius == -1:
        # Disables mask, here to reduce code duplication.
        return np.full((L, L), fill_value=True, dtype=bool)

    elif not 0 < support_radius <= L // 2 * np.sqrt(2):
        raise ValueError(
            "support_radius should be"
            f" `(0, L*sqrt(2)={L*np.sqrt(2)}]` or -1 to disable."
            f" passed {support_radius}."
        )

    g2d = grid_2d(L, indexing="yx", normalized=False, dtype=dtype)
    mask = g2d["r"] < support_radius

    return mask



"""
Utilties for arrays/n-dimensional matrices.
"""

import numpy as np
from scipy.linalg import eigh

from aspire.utils.matlab_compat import m_reshape

SQRT2 = np.sqrt(2)
SQRT2_R = 1 / SQRT2


def unroll_dim(X, dim):
    # TODO: dim is still 1-indexed like in MATLAB to reduce headaches for now
    # TODO: unroll/roll are great candidates for a context manager since they're always used in conjunction.
    dim = dim - 1
    old_shape = X.shape
    new_shape = old_shape[:dim]

    new_shape += (-1,)

    Y = m_reshape(X, new_shape)

    removed_dims = old_shape[dim:]

    return Y, removed_dims


def roll_dim(X, dim):
    # TODO: dim is still 1-indexed like in MATLAB to reduce headaches for now
    old_shape = X.shape
    new_shape = old_shape[:-1] + dim
    Y = m_reshape(X, new_shape)
    return Y


def im_to_vec(im):
    """
    Roll up images into vectors

    :param im: An N-by-N-by-... array.
    :return: An N^2-by-... array.
    """
    shape = im.shape
    assert im.ndim >= 2, "Array should have at least 2 dimensions"
    assert shape[0] == shape[1], "Array should have first 2 dimensions identical"

    return m_reshape(im, (shape[0] ** 2,) + (shape[2:]))


def vol_to_vec(X):
    """
    Roll up volumes into vectors

    :param X: N-by-N-by-N-by-... array.
    :return: An N^3-by-... array.
    """
    shape = X.shape
    assert X.ndim >= 3, "Array should have at least 3 dimensions"
    assert (
        shape[0] == shape[1] == shape[2]
    ), "Array should have first 3 dimensions identical"

    return m_reshape(X, (shape[0] ** 3,) + (shape[3:]))


def vec_to_im(X):
    """
    Unroll vectors to images

    :param X: N^2-by-... array.
    :return: An N-by-N-by-... array.
    """
    shape = X.shape
    N = round(shape[0] ** (1 / 2))
    assert N**2 == shape[0], "First dimension of X must be square"

    return m_reshape(X, (N, N) + (shape[1:]))


def vec_to_vol(X):
    """
    Unroll vectors to volumes

    :param X: N^3-by-... array.
    :return: An N-by-N-by-N-by-... array.
    """
    shape = X.shape
    N = round(shape[0] ** (1 / 3))
    assert N**3 == shape[0], "First dimension of X must be cubic"

    return m_reshape(X, (N, N, N) + (shape[1:]))


def vecmat_to_volmat(X):
    """
    Roll up vector matrices into volume matrices

    :param X: A vector matrix of size L1^3-by-L2^3-by-...
    :return: A volume "matrix" of size L1-by-L1-by-L1-by-L2-by-L2-by-L2-by-...
    """
    # TODO: Use context manager?
    shape = X.shape
    assert X.ndim >= 2, "Array should have at least 2 dimensions"

    L1 = round(shape[0] ** (1 / 3))
    L2 = round(shape[1] ** (1 / 3))

    assert L1**3 == shape[0], "First dimension of X must be cubic"
    assert L2**3 == shape[1], "Second dimension of X must be cubic"

    return m_reshape(X, (L1, L1, L1, L2, L2, L2) + (shape[2:]))


def volmat_to_vecmat(X):
    """
    Unroll volume matrices to vector matrices

    :param X: A volume "matrix" of size L1-by-L1-by-L1-by-L2-by-L2-by-L2-by-...
    :return: A vector matrix of size L1^3-by-L2^3-by-...
    """
    # TODO: Use context manager?
    shape = X.shape
    assert X.ndim >= 6, "Array should have at least 6 dimensions"
    assert shape[0] == shape[1] == shape[2], "Dimensions 1-3 should be identical"
    assert shape[3] == shape[4] == shape[5], "Dimensions 4-6 should be identical"

    l1 = shape[0]
    l2 = shape[3]

    return m_reshape(X, (l1**3, l2**3) + (shape[6:]))


def mdim_mat_fun_conj(X, d1, d2, f):
    """
    Conjugate a multidimensional matrix using a linear mapping

    :param X: An N_1-by-...-by-N_d1-by-N_1...-by-N_d1-by-... array, with the first 2*d1 dimensions corresponding to
        matrices with columns and rows of dimension d1.
    :param d1: The dimension of the input matrix X
    :param d2: The dimension of the output matrix Y
    :param f: A function handle of a linear map that takes an array of size N_1-by-...-by-N_d1-by-... and returns an
        array of size M_1-by-...-by-M_d2-by-... .
    :return: An array of size M_1-by-...-by-M_d2-by-M_1-by-...-by-M_d2-by-... resulting from applying fun to the rows
        and columns of the multidimensional matrix X.
    """

    # Roll up outer dimensions if any.
    dim = 2 * d1 + 1
    sz_roll = X.shape[:-dim]
    shp = X.shape[-dim:]
    X = X.reshape((-1, *shp))

    X = f(X)

    # Swap the last d2 axes with the first d1 axes
    dims1 = [X.ndim - d - 1 for d in range(d1 + d2)]
    dims2 = [X.ndim - d - 1 for d in list(range(d1, d1 + d2)) + list(range(d1))]
    X = np.moveaxis(X, dims1, dims2)

    X = np.conj(X)
    X = f(X)

    # Swap the next d2 axes block
    dims1 = [X.ndim - d - 1 for d in range(2 * d2)]
    dims2 = [X.ndim - d - 1 for d in list(range(d2, 2 * d2)) + list(range(d2))]
    X = np.moveaxis(X, dims1, dims2)

    X = np.conj(X)

    # Unroll outer dimensions.
    X = X.reshape(*sz_roll, *X.shape[1:])

    return X


def symmat_to_vec_iso(mat):
    """
    Isometrically maps a symmetric matrix to a packed vector

    :param mat: An array of size N-by-N-by-... where the first two dimensions constitute symmetric or Hermitian
        matrices.
    :return: A vector of size N*(N+1)/2-by-... consisting of the lower triangular part of each matrix, reweighted so
        that the Frobenius inner product is mapped to the Euclidean inner product.
    """
    mat, sz_roll = unroll_dim(mat, 3)
    N = mat.shape[0]
    mat = mat_to_vec(mat)
    mat[np.arange(0, N**2, N + 1)] *= SQRT2_R
    mat *= SQRT2
    mat = vec_to_mat(mat)
    mat = roll_dim(mat, sz_roll)
    vec = symmat_to_vec(mat)

    return vec


def vec_to_symmat_iso(vec):
    """
    Isometrically map packed vector to symmetric matrix

    :param vec: A vector of size N*(N+1)/2-by-... describing a symmetric (or Hermitian) matrix.
    :return: An array of size N-by-N-by-... which indexes symmetric/Hermitian matrices that occupy the first two
        dimensions. The lower triangular parts of these matrices consists of the corresponding vectors in vec,
        reweighted so that the Euclidean inner product maps to the Frobenius inner product.
    """
    mat = vec_to_symmat(vec)
    mat, sz_roll = unroll_dim(mat, 3)
    N = mat.shape[0]
    mat = mat_to_vec(mat)
    mat[np.arange(0, N**2, N + 1)] *= SQRT2
    mat *= SQRT2_R
    mat = vec_to_mat(mat)
    mat = roll_dim(mat, sz_roll)
    return mat


def symmat_to_vec(mat):
    """
    Packs a symmetric matrix into a lower triangular vector

    :param mat: An array of size N-by-N-by-... where the first two dimensions constitute symmetric or
        Hermitian matrices.
    :return: A vector of size N*(N+1)/2-by-... consisting of the lower triangular part of each matrix.

    Note that a lot of acrobatics happening here (swapaxes/triu instead of tril etc.) are so that we can get
    column-major ordering of elements (to get behavior consistent with MATLAB), since masking in numpy only returns
    data in row-major order.
    """
    N = mat.shape[0]
    assert mat.shape[1] == N, "Matrix must be square"

    mat, sz_roll = unroll_dim(mat, 3)
    triu_indices = np.triu_indices(N)
    vec = mat.swapaxes(0, 1)[triu_indices]
    vec = roll_dim(vec, sz_roll)

    return vec


def vec_to_symmat(vec):
    """
    Convert packed lower triangular vector to symmetric matrix

    :param vec: A vector of size N*(N+1)/2-by-... describing a symmetric (or Hermitian) matrix.
    :return: An array of size N-by-N-by-... which indexes symmetric/Hermitian matrices that occupy the first two
        dimensions. The lower triangular parts of these matrices consists of the corresponding vectors in vec.
    """
    # TODO: Handle complex values in vec
    if np.iscomplex(vec).any():
        raise NotImplementedError("Coming soon")

    # M represents N(N+1)/2
    M = vec.shape[0]
    N = int(round(np.sqrt(2 * M + 0.25) - 0.5))
    assert (
        M == 0.5 * N * (N + 1)
    ) and N != 0, "Vector must be of size N*(N+1)/2 for some N>0."

    vec, sz_roll = unroll_dim(vec, 2)
    index_matrix = np.empty((N, N))
    i_upper = np.triu_indices_from(index_matrix)
    index_matrix[i_upper] = np.arange(
        M
    )  # Incrementally populate upper triangle in row major order
    index_matrix.T[i_upper] = index_matrix[i_upper]  # Copy to lower triangle

    mat = vec[index_matrix.flatten("F").astype("int")]
    mat = m_reshape(mat, (N, N) + mat.shape[1:])
    mat = roll_dim(mat, sz_roll)

    return mat


def mat_to_vec(mat, is_symmat=False):
    """
    Converts a matrix into vectorized form

    :param mat: An array of size N-by-N-by-... containing the matrices to be vectorized.
    :param is_symmat: Specifies whether the matrices are symmetric/Hermitian, in which case they are stored in packed
        form using symmat_to_vec (default False).
    :return: The vectorized form of the matrices, with dimension N^2-by-... or N*(N+1)/2-by-... depending on the value
        of is_symmat.
    """
    if not is_symmat:
        sz = mat.shape
        N = sz[0]
        assert sz[1] == N, "Matrix must be square"
        return m_reshape(mat, (N**2,) + sz[2:])
    else:
        return symmat_to_vec(mat)


def vec_to_mat(vec, is_symmat=False):
    """
    Converts a vectorized matrix into a matrix

    :param vec: The vectorized representations. If the matrix is non-symmetric, this array has the dimensions
        N^2-by-..., but if the matrix is symmetric, the dimensions are N*(N+1)/2-by-... .
    :param is_symmat: True if the vectors represent symmetric matrices (default False)
    :return: The array of size N-by-N-by-... representing the matrices.
    """
    if not is_symmat:
        sz = vec.shape
        N = int(round(np.sqrt(sz[0])))
        assert sz[0] == N**2, "Vector must represent square matrix."
        return m_reshape(vec, (N, N) + sz[1:])
    else:
        return vec_to_symmat(vec)


def make_symmat(A):
    """
    Symmetrize a matrix

    :param A: A matrix.
    :return: The Hermitian matrix (A+A')/2.
    """
    return 0.5 * (A + A.T)


def make_psd(A):
    """
    Make a matrix positive semi-definite

    This is the simplest way by setting negative eigenvalues to zero.

    :param A: A matrix.
    :return: The positive semi-definite matrix
    """
    B = make_symmat(A)
    W, V = eigh(B)
    W[W < 0.0] = 0.0

    return V @ np.diag(W) @ np.transpose(V)


def anorm(x, axes=None):
    """
    Calculate array norm along given axes

    :param x: An array of arbitrary size and shape.
    :param axes: The axis along which to compute the norm. If None, the norm is calculated along all axes.
    :return: The Euclidean (l^2) norm of x along specified axes.
    """
    if axes is None:
        norm = np.linalg.norm(x)
    else:
        axes = tuple(axes)  # Unrolls any generators, like `range`.
        norm = np.sqrt(ainner(x, x, axes=axes))
    return norm


def acorr(x, y, axes=None):
    """
    Calculate array correlation along given axes

    :param x: An array of arbitrary shape
    :param y: An array of same shape as x
    :param axes: The axis along which to compute the correlation. If None, the correlation is calculated along all axes.
    :return: The correlation of x along specified axes.
    """
    assert x.shape == y.shape, "The shapes of the inputs have to match"

    if axes is None:
        axes = range(x.ndim)
    return ainner(x, y, axes) / (anorm(x, axes) * anorm(y, axes))


def ainner(x, y, axes=None):
    """
    Calculate array inner product along given axes

    :param x: An array of arbitrary shape
    :param y: An array of same shape as x
    :param axes: The axis along which to compute the inner product. If None, the product is calculated along all axes.
    :return:
    """
    assert x.shape == y.shape, "The shapes of the inputs have to match"

    if axes is not None:
        axes = tuple(axes)  # Unrolls any generators, like `range`.

    return np.sum(x * np.conj(y), axis=axes)


def eigs(A, k):
    """
    Multidimensional partial eigendecomposition

    :param A: An array of size `sig_sz`-by-`sig_sz`, where `sig_sz` is a size containing d dimensions.
        The array represents a matrix with d indices for its rows and columns.
    :param k: The number of eigenvalues and eigenvectors to calculate (default 6).
    :return: A 2-tuple of values
        V: An array of eigenvectors of size `sig_sz`-by-k.
        D: A matrix of size k-by-k containing the corresponding eigenvalues in the diagonals.
    """
    sig_sz = A.shape[: int(A.ndim / 2)]
    sig_len = np.prod(sig_sz)
    A = m_reshape(A, (sig_len, sig_len))

    dtype = A.dtype
    w, v = eigh(A.astype(np.float64), eigvals=(sig_len - 1 - k + 1, sig_len - 1))

    # Arrange in descending order (flip column order in eigenvector matrix) and typecast to proper type
    w = w[::-1].astype(dtype)
    v = np.fliplr(v)

    v = m_reshape(v, sig_sz + (k,)).astype(dtype)

    return v, np.diag(w)


def best_rank1_approximation(A):
    """
    Computes the best rank-1 approximation of A.

    :param A: A 2D array or a 3D array where the first axis is the stack axis.
    :return: rank-1 ndarray of same size.
    """
    og_shape = A.shape

    if A.ndim == 2:
        A = A[np.newaxis]
    assert A.ndim == 3, "Array must be 2D or 3D representing a stack."

    U, S, V = np.linalg.svd(A)
    S_rank1 = np.zeros_like(A)
    S_rank1[:, 0, 0] = S[:, 0]

    return (U @ S_rank1 @ V).reshape(og_shape)


def fix_signs(u):
    """
    Negates columns so the sign of the largest element in the column is positive.

    For complex values sign is taken as norm(x)/x, zero columns unchanged.

    Typically this is used for making eigenvectors deterministically signed.

    :param u: matrix as numpy array
    :return: matrix as numpy array
    """

    # Locate the largest element in each column
    # Internally np.absolute performs `norm` for complex values.
    index_array = np.argmax(np.absolute(u), axis=0)

    # Create array of sign corrections
    signs = np.take_along_axis(u, np.expand_dims(index_array, axis=0), axis=0).squeeze()
    _abs = np.absolute(signs)
    signs = np.divide(_abs, signs, where=_abs != 0)

    # Now we only care about the sign +1/-1.
    #  The following corrects for any numerical division noise,
    #  and also remaps 0 to +1.
    signs = np.sign(signs * 2 + 1)

    # Apply signs elementwise to matrix
    return u * signs



import logging

import numpy as np
from numpy.linalg import norm

from aspire.abinitio import CLSymmetryC3C4
from aspire.utils import (
    J_conjugate,
    Rotation,
    all_pairs,
    anorm,
    best_rank1_approximation,
    cyclic_rotations,
    tqdm,
    trange,
)
from aspire.utils.random import Random, randn

logger = logging.getLogger(__name__)


class CLSymmetryCn(CLSymmetryC3C4):
    """
    Define a class to estimate 3D orientations using common lines methods for molecules with
    Cn cyclic symmetry, with n>4.
    """

    def __init__(
        self,
        src,
        symmetry=None,
        n_rad=None,
        n_theta=None,
        max_shift=0.15,
        shift_step=1,
        epsilon=1e-3,
        max_iters=1000,
        degree_res=1,
        n_points_sphere=500,
        equator_threshold=10,
        seed=None,
    ):
        """
        Initialize object for estimating 3D orientations for molecules with Cn symmetry, n>4.

        :param src: The source object of 2D denoised or class-averaged images with metadata
        :param symmetry: A string, 'Cn', indicating the symmetry type.
        :param n_rad: The number of points in the radial direction.
        :param n_theta: The number of points in the theta direction.
        :param max_shift: Maximum range for shifts as a proportion of resolution. Default = 0.15.
        :param shift_step: Resolution of shift estimation in pixels. Default = 1 pixel.
        :param epsilon: Tolerance for the power method.
        :param max_iter: Maximum iterations for the power method.
        :param degree_res: Degree resolution for estimating in-plane rotations.
        :param n_points_sphere: The number of candidate rotations used to estimate viewing directions.
        :param equator_threshold: Threshold for removing candidate rotations within `equator_threshold`
            degrees of being an equator image. Default is 10 degrees.
        :param seed: Optional seed for RNG.
        """

        super().__init__(
            src,
            symmetry=symmetry,
            n_rad=n_rad,
            n_theta=n_theta,
            max_shift=max_shift,
            shift_step=shift_step,
            epsilon=epsilon,
            max_iters=max_iters,
            degree_res=degree_res,
            seed=seed,
        )

        self.n_points_sphere = n_points_sphere
        self.equator_threshold = equator_threshold

    def _check_symmetry(self, symmetry):
        if symmetry is None:
            raise NotImplementedError(
                "Symmetry type not supplied. Please indicate symmetry."
            )
        else:
            symmetry = symmetry.upper()
            _sym_type = symmetry[0]
            _order = int(symmetry[1:])
            if not (_sym_type == "C" and _order > 4):
                raise NotImplementedError(
                    f"Only Cn symmetry, n > 4, supported. {symmetry} was supplied."
                )
            self.order = _order

    def estimate_rotations(self):
        """
        Estimate rotation matrices for molecules with Cn symmetry, n > 4.

        :return: Array of rotation matrices, size n_imgx3x3.
        """
        super().estimate_rotations()

    def _estimate_relative_viewing_directions(self):
        logger.info(f"Estimating relative viewing directions for {self.n_img} images.")
        pf = self.pf

        # Generate candidate rotation matrices and the common-line and
        # self-common-line indices induced by those rotations.
        Ris_tilde, R_theta_ijs = self.generate_candidate_rots(
            self.n_points_sphere,
            self.equator_threshold,
            self.order,
            self.degree_res,
            self.seed,
        )
        cijs_inds = self._compute_cls_inds(Ris_tilde, R_theta_ijs)
        scls_inds = self._compute_scls_inds(Ris_tilde)
        n_theta_ijs = len(R_theta_ijs)

        # Generate shift phases.
        r_max = pf.shape[-1]
        shifts, shift_phases, _ = self._generate_shift_phase_and_filter(
            r_max, self.max_shift, self.shift_step
        )
        n_shifts = len(shifts)
        all_shift_phases = shift_phases.T

        # Reconstruct full polar Fourier for use in correlation.
        pf /= norm(pf, axis=2)[..., np.newaxis]  # Normalize each ray.
        pf_full = np.concatenate((pf, np.conj(pf)), axis=1)

        # Pre-compute shifted pf's.
        pf_shifted = (pf * all_shift_phases[:, None, None]).swapaxes(0, 1)
        pf_shifted = pf_shifted.reshape(
            (self.n_img, n_shifts * (self.n_theta // 2), r_max)
        )

        # Step 1: pre-calculate the likelihood with respect to the self-common-lines.
        scores_self_corrs = self._scl_likelihood(pf_shifted, pf_full, scls_inds)

        # Step 2: Compute the likelihood for each pair of candidate matrices with respect
        # to the common-lines they induce. Then compute estimates for viis and vijs.
        logger.info("Computing pairwise likelihood.")
        n_vijs = self.n_img * (self.n_img - 1) // 2
        vijs = np.zeros((n_vijs, 3, 3), dtype=self.dtype)
        viis = np.zeros((self.n_img, 3, 3), dtype=self.dtype)
        rots_symm = cyclic_rotations(self.order, self.dtype).matrices

        # List of MeanOuterProductEstimator instances.
        # Used to keep a running mean of J-synchronized estimates for vii.
        mean_est = []
        for _ in range(self.n_img):
            mean_est.append(MeanOuterProductEstimator())

        pairs = all_pairs(self.n_img)
        for ind, (i, j) in enumerate(tqdm(pairs)):
            # Compute correlation.
            corrs_ij = np.real(pf_shifted[i] @ np.conj(pf_full[j]).T)

            # Max out over shifts.
            corrs_ij = np.max(
                np.reshape(corrs_ij, (n_shifts, self.n_theta // 2, self.n_theta)),
                axis=0,
            )

            # Arrange correlation based on common lines induced by candidate rotations.
            corrs = corrs_ij[cijs_inds[..., 0], cijs_inds[..., 1]]
            corrs = np.reshape(
                corrs,
                (
                    self.n_points_sphere,
                    self.n_points_sphere,
                    self.order,
                    n_theta_ijs // self.order,
                ),
            )

            # Take the mean over all symmetry induced common lines.
            corrs = np.mean(corrs, axis=-2)

            # Compute maximum likelihood while taking into consideration both cls and scls.
            # Get indices of optimal candidates for Ri_tilde, Rj_tilde, and R_theta_ij.
            corrs = (
                corrs
                * np.outer(scores_self_corrs[i], scores_self_corrs[j])[..., np.newaxis]
            )
            opt_i, opt_j, opt_ij = np.unravel_index(np.argmax(corrs), corrs.shape)

            # Optimal candidate rotations.
            opt_Ri_tilde = Ris_tilde[opt_i]
            opt_Rj_tilde = Ris_tilde[opt_j]
            opt_R_theta_ij = R_theta_ijs[opt_ij]

            # Compute the estimate of vi*vi.T as given by j.
            vii_j = np.mean(opt_Ri_tilde.T @ rots_symm @ opt_Ri_tilde, axis=0)
            mean_est[i].push(vii_j)

            # Compute the estimate of vj*vj.T as given by i.
            vjj_i = np.mean(opt_Rj_tilde.T @ rots_symm @ opt_Rj_tilde, axis=0)
            mean_est[j].push(vjj_i)

            # Compute the estimate of vi*vj.T.
            vijs[ind] = np.mean(
                opt_Ri_tilde.T @ rots_symm @ opt_R_theta_ij @ opt_Rj_tilde,
                axis=0,
            )

        # There are conflicting methods betweeen the paper and the Matlab code for
        # finding the optimal estimates for viis. The paper suggests using SVD to find
        # the estimate for vii which is closest to rank 1. The Matlab code takes the
        # median over the stack of all estimates for vii. Here we have implemented
        # a method which J-synchronizes the estimates prior to taking the mean.
        # See issue #869 for more details.
        for i in range(self.n_img):
            viis[i] = mean_est[i].synchronized_mean()

        # As we are using a mean to get the estimates, viis, the estimate will not be rank-1
        # So we use SVD to find a close rank-1 approximation.
        viis_rank1 = best_rank1_approximation(viis)

        return vijs, viis_rank1

    def _scl_likelihood(self, pf_shifted, pf_full, scls_inds):
        scores_self_corrs = np.zeros((self.n_img, len(scls_inds)), dtype=self.dtype)
        logger.info("Computing likelihood wrt self common-lines.")
        for i in trange(self.n_img):
            # Compute correlation of pf[i] with itself over all shifts.
            corrs = np.real(pf_shifted[i] @ np.conj(pf_full[i]).T)
            # Reshape to (n_shifts, n_theta//2, n_theta).
            corrs = np.reshape(corrs, (-1, self.n_theta // 2, self.n_theta))
            corrs_cands = np.max(
                (corrs[:, scls_inds[:, :, 0], scls_inds[:, :, 1]]), axis=0
            )

            scores_self_corrs[i] = np.mean(corrs_cands, axis=1)

        return scores_self_corrs

    def _compute_scls_inds(self, Ris_tilde):
        """
        Compute self-common-lines indices induced by candidate rotations.

        :param Ris_tilde: An array of size n_candsx3x3 of candidate rotations.
        :return: An n_cands x (order-1)//2 x 2 array holding the indices of the (order-1)//2
            non-collinear pairs of self-common-lines for each candidate rotation.
        """
        n_scl_pairs = (self.order - 1) // 2
        n_cands = Ris_tilde.shape[0]
        scls_inds = np.zeros((n_cands, n_scl_pairs, 2), dtype=np.uint16)
        rots_symm = cyclic_rotations(self.order, dtype=self.dtype).matrices

        for i_cand, Ri_cand in enumerate(Ris_tilde):
            Riigs = Ri_cand.T @ rots_symm[1 : n_scl_pairs + 1] @ Ri_cand

            c1s, c2s = self.relative_rots_to_cl_indices(Riigs, self.n_theta)

            scls_inds[i_cand, :, 0] = c1s
            scls_inds[i_cand, :, 1] = c2s
        return scls_inds

    def _compute_cls_inds(self, Ris_tilde, R_theta_ijs):
        """
        Compute the common-lines indices induced by the candidate rotations.

        :param Ris_tilde: An array of size n_candsx3x3 of candidate rotations.
        :param R_theta_ijs: An array of size n_theta_ijsx3x3 of inplane rotations.
        :return: An array of size n_cands x n_cands x n_theta_ijs x 2 holding common-lines
            indices induced by the supplied rotations.
        """
        n_points_sphere = self.n_points_sphere
        n_theta_ijs = R_theta_ijs.shape[0]
        cij_inds = np.zeros(
            (n_points_sphere, n_points_sphere, n_theta_ijs, 2), dtype=np.uint16
        )
        logger.info("Computing common-line indices induced by candidate rotations.")
        for i in tqdm(range(n_points_sphere)):
            for j in range(n_points_sphere):
                R_cands = Ris_tilde[i].T @ R_theta_ijs @ Ris_tilde[j]

                c1s, c2s = self.relative_rots_to_cl_indices(R_cands, self.n_theta)

                cij_inds[i, j, :, 0] = c1s
                cij_inds[i, j, :, 1] = c2s
        return cij_inds

    @staticmethod
    def relative_rots_to_cl_indices(relative_rots, n_theta):
        """
        Given a set of relative rotations between pairs of images
        produce the common-line indices for each pair.

        :param relative_rots: The n x 3 x 3 relative rotations between pairs of images.
        :param n_theta: The theta resolution for common-line indices.

        :return: Common-line indices c1s, c2s each length n.
        """
        c1s = np.array((-relative_rots[:, 1, 2], relative_rots[:, 0, 2])).T
        c2s = np.array((relative_rots[:, 2, 1], -relative_rots[:, 2, 0])).T

        c1s = CLSymmetryC3C4.cl_angles_to_ind(c1s, n_theta)
        c2s = CLSymmetryC3C4.cl_angles_to_ind(c2s, n_theta)

        inds = np.where(c1s >= n_theta // 2)
        c1s[inds] -= n_theta // 2
        c2s[inds] += n_theta // 2
        c2s[inds] = np.mod(c2s[inds], n_theta)

        return c1s, c2s

    @staticmethod
    def generate_candidate_rots(n, equator_threshold, order, degree_res, seed):
        """
        Generate random rotations that exclude rotations inducing equator images
        for use as candidates in the CLSymmetryCn algorithm.

        :param n: Number of rotations to generate.
        :param equator_threshold: Angular distance from equator (in degrees).
        :param order: Cyclic order of underlying molecule.
        :param degree_res: Degree resolution for in-plane rotations.
        :param seed: Random seed.

        :returns: Candidate rotations, In-plane rotations
        """

        logger.info("Generating candidate rotations.")
        # Construct candidate rotations, Ris_tilde.
        Ris_tilde = np.zeros((n, 3, 3))
        counter = 0
        with Random(seed):
            while counter < n:
                third_row = randn(3)
                third_row /= anorm(third_row, axes=(-1,))
                Ri_tilde = CLSymmetryC3C4._complete_third_row_to_rot(third_row)

                # Exclude candidates that represent equator images. Equator candidates
                # induce collinear self-common-lines, which always have perfect correlation.
                angle_from_equator = abs(np.arccos(Ri_tilde[2, 2]) - np.pi / 2)
                if angle_from_equator >= equator_threshold * np.pi / 180:
                    Ris_tilde[counter] = Ri_tilde
                    counter += 1

        # Construct all in-plane rotations, R_theta_ijs
        # The number of R_theta_ijs must be divisible by the symmetric order.
        n_theta_ij = 360 - (360 % order)
        theta_ij = np.arange(0, n_theta_ij, degree_res) * np.pi / 180
        R_theta_ijs = Rotation.about_axis("z", theta_ij).matrices

        return Ris_tilde, R_theta_ijs


class MeanOuterProductEstimator:
    """
    Incrementally accumulate outer product entries of unknown conjugation.
    """

    # These arrays are small enough to just use doubles.
    # Then we can probably avoid numerical summing concerns without precomputing denom
    dtype = np.float64

    def __init__(self):
        # Create storage for J-synchronized estimates.
        self.V_estimates_sync = np.zeros((3, 3), dtype=self.dtype)
        self.count = 0

    def push(self, V):
        """
        Given V, accumulate entries into a running sum of J-synchronized entries.
        """

        # Accumulate synchronized entries to compute synchronized mean.
        # Method added Feb, 2023. (Josh Carmichael)
        if self.count == 0:
            self.V_estimates_sync += V
        else:
            if (np.sign(V) == np.sign(self.V_estimates_sync)).all():
                self.V_estimates_sync += V
            else:
                self.V_estimates_sync += J_conjugate(V)

        self.count += 1

    def synchronized_mean(self):
        """
        Calculate the mean of synchronized outer product estimates.
        """
        return self.V_estimates_sync / self.count
